# archive-pdf-tools
# Copyright (C) 2020-2021, Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Merlijn Boris Wolf Wajer <merlijn@archive.org>

VERSION = '1.5.2'

PRODUCER = 'Internet Archive PDF %s; including '\
           'mupdf and pymupdf/skimage' % (VERSION,)


IMAGE_MODE_PASSTHROUGH = 0
IMAGE_MODE_PIXMAP = 1
IMAGE_MODE_MRC = 2
IMAGE_MODE_SKIP = 3


DENOISE_NONE = 'none'
DENOISE_FAST = 'fast'
DENOISE_BREGMAN = 'bregman'

RECODE_RUNTIME_WARNING_INVALID_PAGE_SIZE = 'invalid-page-size'
RECODE_RUNTIME_WARNING_INVALID_PAGE_NUMBERS = 'invalid-page-numbers'
RECODE_RUNTIME_WARNING_INVALID_JP2_HEADERS = 'invalid-jp2-headers'
RECODE_RUNTIME_WARNING_TOO_SMALL_TO_DOWNSAMPLE = 'too-small-to-downsample'

RECODE_RUNTIME_WARNINGS = {
    RECODE_RUNTIME_WARNING_INVALID_PAGE_SIZE,
    RECODE_RUNTIME_WARNING_INVALID_PAGE_NUMBERS,
    RECODE_RUNTIME_WARNING_INVALID_JP2_HEADERS,
    RECODE_RUNTIME_WARNING_TOO_SMALL_TO_DOWNSAMPLE,
}

JPEG2000_IMPL_KAKADU = 'kakadu'
JPEG2000_IMPL_OPENJPEG = 'openjpeg'
JPEG2000_IMPL_GROK = 'grok'
JPEG2000_IMPL_PILLOW = 'pillow'

JPEG2000_IMPLS = (JPEG2000_IMPL_KAKADU, JPEG2000_IMPL_OPENJPEG,
        JPEG2000_IMPL_GROK, JPEG2000_IMPL_PILLOW)

COMPRESSOR_JPEG2000 = 'jpeg2000'
COMPRESSOR_JPEG = 'jpeg'

COMPRESSOR_JBIG2 = 'jbig2'
COMPRESSOR_CCITT = 'ccitt'

__version__ = VERSION



# archive-pdf-tools
# Copyright (C) 2020-2021, Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Merlijn Boris Wolf Wajer <merlijn@archive.org>


soorten profiles:

- profile voor mrc image *generatie*: dpi, downsample, denoise

- profile voor compressie (jpeg,jpeg2000,compression params, jbig2,ccitt), hq
  pages, image_mode


# TODO: Use argparse groups for these

image_args = [
    args.denoise_mask,
    args.downsample
    args.bg_downsample
    args.fg_downsample,
    args.dpi,
]

compression_args = [
    args.image_mode,
    args.mask_compression == COMPRESSOR_JBIG2,
    args.jpeg2000_implementation,
    args.bg_compression_flags.split(' '),
    args.fg_compression_flags.split(' '),
    args.mrc_image_format,
    args.hq_pages,
    args.hq_bg_compression_flags.split(' '),
    args.hq_fg_compression_flags.split(' '),
]

metadata_args = [
    args.metadata_url, args.metadata_title, args.metadata_author,
    args.metadata_creator, args.metadata_language,
    args.metadata_subject, args.metadata_creatortool,
]

input_args = [
    args.from_pdf, args.from_imagestack,
    args.hocr_file, args.scandata_file,
    args.out_pdf,
    args.out_dir,
]

misc_args = [
    args.reporter,
    args.verbose,
    args.report_every,
    args.stop_after,
    args.render_text_lines,
    args.tmp_dir,

    args.grayscale_pdf,
    args.bw_pdf,
]





# archive-pdf-tools
# Copyright (C) 2020-2021, Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Merlijn Boris Wolf Wajer <merlijn@archive.org>

import re

# TODO:
# - Implement page number prefix
# - Deal with ambiguity between roman numerals 'I', 'II', 'MM' and Alpha 'I',
#   'II', etc.

from roman import fromRoman, InvalidRomanNumeralError

INVALID, ARABIC, ROMAN_LOWER, ROMAN_UPPER, ALPHA_UPPER, ALPHA_LOWER = range(6)

_type2str = {
        INVALID: 'Invalid',
        ARABIC: 'Arabic',
        ROMAN_LOWER: 'Roman lower',
        ROMAN_UPPER: 'Roman upper',
        ALPHA_UPPER: 'Alpha upper',
        ALPHA_LOWER: 'Alpha lower'
}

_type2pdf = {
        ARABIC: '/D',
        ROMAN_LOWER: '/r',
        ROMAN_UPPER: '/R',
        ALPHA_UPPER: '/A',
        ALPHA_LOWER: '/a',
}

_ARABIC_RE = re.compile('[0-9]+')
_ALPHA_UPPER_RE = re.compile('[A-Z]+')
_ALPHA_LOWER_RE = re.compile('[a-z]+')


def is_arabic(n):
    return _ARABIC_RE.match(n) and n.isnumeric()


def is_alpha_upper(n):
    return _ALPHA_UPPER_RE.match(n)


def is_alpha_lower(n):
    return _ALPHA_LOWER_RE.match(n)


def is_roman(n):
    try:
        _ = fromRoman(n.upper())
        return True
    except InvalidRomanNumeralError:
        return False


def is_roman_lower(n):
    return n.lower() == n and is_roman(n)


def is_roman_upper(n):
    return n.upper() == n and is_roman(n)


# Evince goes from 'Z' to 'AA' to 'ZZ' to 'AAA'.
def alpha_to_number(n):
    # 'A' = 1
    # 'Z' = 26
    # 'AA' = 27
    # 'ZZ' = 52
    # 'AAA' = 53

    first = True
    res = 1
    for c in n:
        tmp = ord(c) - ord('A')
        res += tmp

        if first:
            first = False
        else:
            res += 26 - tmp

    return res


def get_val_type(v):
    if v is None:
        # XXX: Does this special case make sense?
        return INVALID
    elif is_arabic(v):
        return ARABIC
    # Prefer roman over alpha since we do not yet support alpha
    elif is_roman_lower(v):
        return ROMAN_LOWER
    elif is_roman_upper(v):
        return ROMAN_UPPER
    elif is_alpha_upper(v):
        return ALPHA_UPPER
    elif is_alpha_lower(v):
        return ALPHA_LOWER
    else:
        raise ValueError('Page number not in spec: %s' % repr(v))


def get_val_value(v, vtype):
    if v and ' ' in v:
        v = v.strip().split(' ')[0]
    if vtype == INVALID:
        return None
    elif vtype == ARABIC:
        return int(v, 10)
    elif vtype == ROMAN_LOWER or vtype == ROMAN_UPPER:
        try:
            return fromRoman(v.upper())
        except InvalidRomanNumeralError:
            raise ValueError
    elif vtype == ALPHA_LOWER or vtype == ALPHA_UPPER:
        return alpha_to_number(v.upper())
    pass


def find_next_nonzero(series):
    for v in series:
        if v is not None:
            return v
    return None


def parse_series(series):
    # Let's first split up the series:
    # - detect when the numbers are not in sequence
    # - detect when the type changes
    last_value = None
    last_val_type = INVALID

    series_start = 0

    resulting_series = []
    all_ok = True

    running_series = []
    running_series_n = []

    for idx, val in enumerate(series):
        new = False

        try:
            val_type = get_val_type(val)
            val_value = get_val_value(val, val_type)
        except ValueError as e:
            all_ok = False
            val_type = INVALID
            val_value = None

        if val_type in (ROMAN_UPPER, ROMAN_LOWER):
            next_val = find_next_nonzero(series[idx+1:])
            next_val_type = get_val_type(next_val)

            if val_type != next_val_type:
                if val_type == ROMAN_UPPER and next_val_type == ALPHA_UPPER or \
                        val_type == ROMAN_LOWER and next_val_type == ALPHA_LOWER:
                    val_type = next_val_type
                    val_value = get_val_value(val, val_type)

        if val_type in (ALPHA_UPPER, ALPHA_LOWER):
            next_val = find_next_nonzero(series[idx+1:])
            next_val_type = get_val_type(next_val)

            ord_val = None
            ord_next_val = None
            try:
                ord_val = ord(val)
                ord_next_val = ord(next_val)
            except TypeError:
                ord_val = None
                ord_next_val = None

            if next_val is None:
                pass
            elif ord_val is not None and ord_next_val is not None and \
                    ord_val == ord_next_val - 1:
                pass
            elif val_type == next_val_type:
                pass
            elif (val_type == ALPHA_UPPER and next_val_type == ROMAN_UPPER) or \
                 (val_type == ALPHA_LOWER and next_val_type == ROMAN_LOWER):
                     try:
                         val_type = next_val_type
                         val_value = get_val_value(val, val_type)
                     except ValueError:
                         val_type = INVALID
                         val_value = None
            elif val_type in (ALPHA_LOWER, ALPHA_UPPER) and next_val_type not in (ROMAN_UPPER, ROMAN_LOWER):
                # We can have a case where an invalid roman numeral ('XXXVIIII')
                # is followed by a arabic numeral (39), which will hit this
                # case, let's just treat it invalid
                val_type = INVALID
                val_value = None
            else:
                # This code should be unreachable now
                raise ValueError('Alpha page numbers are not supported at the '
                                 ' moment due to ambiguity in the spec.')

        if val_type != last_val_type:
            new = True

        if val_type == INVALID and last_val_type == INVALID:
            pass
        else:
            if last_val_type == INVALID:
                new = True
            elif val_type == INVALID:
                new = True
            elif val_value != last_value + 1:
                new = True

        if new and idx != 0:
            resulting_series.append({'start': series_start,
                                     'type': last_val_type,
                                     'type_human': _type2str[last_val_type],
                                     'values': running_series,
                                     'values_numeric': running_series_n})

            series_start = idx
            running_series = []
            running_series_n = []

        running_series.append(val)
        running_series_n.append(val_value)

        last_value = val_value
        last_val_type = val_type

    resulting_series.append({'start': series_start,
                             'type': last_val_type,
                             'type_human': _type2str[last_val_type],
                             'values': running_series,
                             'values_numeric': running_series_n})

    return resulting_series, all_ok


# https://www.w3.org/TR/WCAG20-TECHS/PDF17.html
# Page labels are specified as follows:
# 
#     /S specifies the numbering style for page numbers:
# 
#         /D - Arabic numerals (1,2,3...)
# 
#         /r - lowercase Roman numerals (i, ii, iii,...)
# 
#         /R - uppercase Roman numerals (I, II, III,...)
# 
#         /A - uppercase letters (A-Z)
# 
#         /a - lowercase letters (a-z)
# 
#     /P (optional) - page number prefix
# 
#     /St (optional) - the value of the first page number in the range (default:
#     1)
def series_to_pdf(series):
    res = '''  /PageLabels <<
    /Nums [ '''

    for s in series:
        r = '%d ' % s['start']
        if s['type'] == INVALID:
            r += '<<\n        >> '
        else:
            r += '<<\n'
            r += '         /S ' + _type2pdf[s['type']] + '\n'
            r += '         /St %d' % s['values_numeric'][0] + '\n'
            r += '        '
            r += '>> '

        res += r

    res += ''']
    >>'''

    return res


if __name__ == '__main__':
    all_series = [
        [None, 'i', 'ii', 'iii', None, None, None, 'iv', 'v', 'v', 'vi', '3',
            '4', '5', '4', '6', 'i', '7', None],
        ['i', 'ii', 'iii', 'vi', '3', '5', '4', '6', 'i', '7', 'A-2', 'B-2'],
        ['', 'i', 'ii',  'vi', '3', '5', '4', '6', 'i', '7', 'A-2', 'B-2'],
        ['i', 'j', 'k', 'l', None, None, None, None, None, None, 'I', 'II', 'III', 'IV', 'V', 'VI', 'VII', 'VIII', 'IX', 'X', 'XI', 'XII', 'XIII', 'XIV', 'XV', 'XVI', 'XVII', 'XVIII', 'XIX', 'XX', 'XXI', None, '565', '566', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23', '24', '25', '26', '27', '28', '29', '30', '31', '32', '33', '34', '35', '36', '37', '38', '39', '40', '41', '42', '43', '44', '45', '46', '47', '48', '49', '50', '51', '52', '53', '54', '55', '56', '57', '58', '59', '60', '61', '62', '63', '64', '65', '66', '67', '68', '69', '70', '71', '72', '73', '74', '75', '76', '77', '78', '79', '80', '81', '82', '83', '84', '85', '86', '87', '88', '89', '90', '91', '92', '93', '94', '95', '96', '97', '98', '99', '100', '101', '102', '103', '104', '105', '106', '107', '108', '109', '110', '111', '112', '113', '114', '115', '116', '117', '118', '119', '120', '121', '122', '123', '124', '125', '126', '127', '128', '129', '130', '131', '132', '133', '134', '135', '136', '137', '138', '139', '140', '141', '142', '143', '144', '145', '146', '147', '148', '149', '150', '151', '152', '153', '154', '155', '156', '157', '158', '159', '160', '161', '162', '163', '164', '165', '166', '167', '168', '169', '170', '171', '172', '173', '174', '175', '176', '177', '178', '179', '180', '181', '182', '183', '184', '185', '186', '187', '188', '189', '190', '191', '192', '193', '194', '195', '196', '197', '198', '199', '200', '201', '202', '203', '204', '205', '206', '207', '208', '209', '210', '211', '212', '213', '214', '215', '216', '217', '218', '219', '220', '221', '222', '223', '224', '225', '226', '227', '228', '229', '230', '231', '232', '233', '234', '235', '236', '237', '238', '239', '240', '241', '242', '243', '244', '245', '246', '247', '248', '249', '250', '251', '252', '253', '254', '255', '256', '257', '258', '259', '260', '261', '262', '263', '264', '265', '266', '267', '268', '269', '270', '271', '272', '273', '274', '275', '276', '277', '278', '279', '280', '281', '282', '283', '284', '285', '286', '287', '288', '289', '290', '291', '292', '293', '294', '295', '296', '297', '298', '299', '300', '301', '302', '303', '304', '305', '306', '307', '308', '309', '310', '311', '312', '313', '314', '315', '316', '317', '318', '319', '320', '321', '322', '323', '324', '325', '326', '327', '328', '329', '330', '331', '332', '333', '334', '335', '336', '337', '338', '339', '340', '341', '342', '343', '344', '345', '346', '347', '348', '349', '350', '351', '352', '353', '354', '355', '356', '357', '358', '359', '360', '361', '362', '363', '364', '365', '366', '367', '368', '369', '370', '371', '372', '373', '374', '375', '376', '377', '378', '379', '380', '381', '382', '383', '384', '385', '386', '387', '388', '389', '390', '391', '392', '393', '394', '395', '396', '397', '398', '399', '400', '401', '402', '403', '404', '405', '406', '407', '408', '409', '410', '411', '412', '413', '414', '415', '416', '417', '418', '419', '420', '421', '422', '423', '424', '425', '426', '427', '428', '429', '430', '431', '432', '433', '434', '435', '436', '437', '438', '439', '440', '441', '442', '443', '444', '445', '446', '447', '448', '449', '450', '451', '452', '453', '454', '455', '456', '457', '458', '459', '460', '461', '462', '463', '464', '465', '466', '467', '468', '469', '470', '471', '472', '473', '474', '475', '476', '477', '478', '479', '480', '481', '482', '483', '484', '485', '486', '487', '488', '489', '490', '491', '492', '493', '494', '495', '496', '497', '498', '499', '500', '501', '502', '503', '504', '505', '506', '507', '508', '509', '510', '511', '512', '513', '514', '515', '516', '517', '518', '519', '520', '521', '522', '523', '524', '525', '526', '527', '528', '529', '530', '531', '532', '533', '534', '535', '536', '537', '538', '539', '540', '541', '542', '543', '544', '545', '546', '547', '548', '549', '550', '551', '552', '553', '554', '555', '556', '557', '558', '559', '560', '561', '562', '563', '564', 'I', None, 'II', None, 'III', None, 'IV', None, 'V', None, 'VI', None, 'VII', None, 'VII A.', None, 'VIII', None, 'IX', None, 'X', None, 'XI', None, 'XII', None, 'XIII', None, 'XIV', None, 'XV', None, 'XVI', None, 'XVII', None, 'XVIII', None, 'XIX', None, 'XX', None, 'XXI', None, 'XXII', None, 'XXIII', None, 'XXIV', None, 'XXV', None, 'XXVI', None, 'XXVII', None, 'XXVIII', None, 'XXIX', None, 'XXX', None, 'XXXI', None, 'XXXII', None, 'XXXIII', None, 'XXXIV', None, None, None, None, None, None, None]
    ]

    for series in all_series:
        res, all_ok = parse_series(series)
        from pprint import pprint
        pprint(res)
        print(all_ok)
        #print(series_to_pdf(res))




"""
  Arabic, Armenian, Bengali, Canadian_Aboriginal, Cherokee, Cyrillic, Devanagari, Ethiopic,
       Fraktur, Georgian, Greek, Gujarati, Gurmukhi, HanS (Han simplified), HanS_vert (Han
       simplified, vertical), HanT (Han traditional), HanT_vert (Han traditional, vertical), Hangul,
       Hangul_vert (Hangul vertical), Hebrew, Japanese, Japanese_vert (Japanese vertical), Kannada,
       Khmer, Lao, Latin, Malayalam, Myanmar, Oriya (Odia), Sinhala, Syriac, Tamil, Telugu, Thaana,
       Thai, Tibetan, Vietnamese.

"""

def visible_font_for_language(lang):
    # TODO: or script?

    # TODO: make sure to keep xref count
    pass


lang2font = {
    ''
}



# archive-pdf-tools
# Copyright (C) 2020-2021, Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Merlijn Boris Wolf Wajer <merlijn@archive.org>
import pkg_resources

import sys
import os
import subprocess
from os import remove
from time import time
from datetime import datetime
from tempfile import mkstemp
from os.path import join
import shutil
import json
from glob import glob
import re
import io


from PIL import Image
import numpy as np
import fitz

from hocr.parse import (hocr_page_iterator, hocr_page_to_word_data,
        hocr_page_get_dimensions, hocr_page_get_scan_res)
from internetarchivepdf.mrc import create_mrc_hocr_components, \
        encode_mrc_images, encode_mrc_mask
from internetarchivepdf.grayconvert import special_gray_convert
from internetarchivepdf.pdfhacks import fast_insert_image, write_pdfa, \
        write_page_labels, write_basic_ua, write_metadata
from internetarchivepdf.pdfrenderer import TessPDFRenderer
from internetarchivepdf.scandata import scandata_xml_get_skip_pages, \
        scandata_xml_get_page_numbers, scandata_xml_get_dpi_per_page, \
        scandata_xml_get_document_dpi
from internetarchivepdf.jpeg2000 import decode_jpeg2000, get_jpeg2000_info
from internetarchivepdf.const import (IMAGE_MODE_PASSTHROUGH, IMAGE_MODE_PIXMAP,
        IMAGE_MODE_MRC, RECODE_RUNTIME_WARNING_INVALID_PAGE_SIZE,
        RECODE_RUNTIME_WARNING_INVALID_PAGE_NUMBERS,
        RECODE_RUNTIME_WARNING_INVALID_JP2_HEADERS, JPEG2000_IMPL_KAKADU,
        JPEG2000_IMPL_OPENJPEG, JPEG2000_IMPL_GROK, JPEG2000_IMPL_PILLOW,
        COMPRESSOR_JPEG2000, COMPRESSOR_JPEG)

PDFA_MIN_UNITS = 3
PDFA_MAX_UNITS = 14400

Image.MAX_IMAGE_PIXELS = 625000000


def guess_dpi(w, h, expected_format=(8.27, 11.69), round_to=[72, 96, 150, 300, 600]):
    """
    Guesstimate DPI for a given image.

    Args:

    * w (int): width of the image
    * h (int): height of the image
    * expected_format (tuple): (width_inch, height_inch) of expected document,
                               defaults to european A4.
    * round_to (list of int): List of acceptable DPI values.
                              Defaults to (72, 96, 150, 300, 600)

    Returns an int which is the best matching DPI picked from round_to.
    """
    w_dpi = w / expected_format[0]
    h_dpi = h / expected_format[1]
    diffs = []
    for dpi in round_to:
        diff = abs(w_dpi - dpi) + abs(h_dpi - dpi)
        diffs.append((dpi, diff))
    sorted_diffs = sorted(diffs, key=lambda x: x[1])
    return sorted_diffs[0][0]


def create_tess_textonly_pdf(hocr_file, save_path, in_pdf=None,
        image_files=None, dpi=None, skip_pages=None, dpi_pages=None,
        reporter=None,
        verbose=False, debug=False, stop_after=None,
        render_text_lines=False,
        tmp_dir=None,
        jpeg2000_implementation=None,
        errors=None):
    hocr_iter = hocr_page_iterator(hocr_file)

    render = TessPDFRenderer(render_text_lines=render_text_lines)
    render.BeginDocumentHandler()

    skipped_pages = 0

    last_time = time()
    reporting_page_count = 0

    if verbose:
        print('Starting page generation at', datetime.utcnow().isoformat())

    for idx, hocr_page in enumerate(hocr_iter):
        w, h = hocr_page_get_dimensions(hocr_page)
        hocr_dpi = hocr_page_get_scan_res(hocr_page)
        # If scan_res is not found in hOCR, it returns (None, None)
        hocr_dpi = hocr_dpi[1]

        if skip_pages is not None and idx in skip_pages:
            if verbose:
                print('Skipping page %d' % idx)
            skipped_pages += 1
            continue

        if stop_after is not None and (idx - skipped_pages) >= stop_after:
            break

        if in_pdf is not None:
            page = in_pdf[idx - skipped_pages]
            width = page.rect.width
            height = page.rect.height

            scaler = page.rect.width / w
            ppi = 72 / scaler
        elif image_files is not None:
            # Do not subtract skipped pages here
            imgfile = image_files[idx]

            if imgfile.endswith('.jp2'):
                size, _ = get_jpeg2000_info(imgfile, jpeg2000_implementation, errors)
                imwidth, imheight = size
            else:
                img = Image.open(imgfile)
                imwidth, imheight = img.size
                del img

            page_dpi = dpi
            per_page_dpi = None

            if dpi_pages is not None:
                try:
                    per_page_dpi = int(dpi_pages[idx - skipped_pages])
                    page_dpi = per_page_dpi
                except:
                    pass  # Keep item-wide dpi if available

            # Both document level dpi is not available and per-page dpi is not
            # available, let's guesstimate
            # Assume european A4 (8.27",11.69") and guess DPI
            # to be one-of (72, 96, 150, 300, 600)
            if page_dpi is None:
                page_dpi = guess_dpi(imwidth, imheight,
                                     expected_format=(8.27, 11.69),
                                     round_to=(72, 96, 150, 300, 600))

            page_width = imwidth / (page_dpi / 72)
            if page_width <= PDFA_MIN_UNITS or page_width >= PDFA_MAX_UNITS:
                if verbose:
                    print('Page size invalid with current image size and dpi.')
                    print('Image size: %d, %d. DPI: %d' % (imwidth, imheight,
                                                           page_dpi))

                # First let's try without per_page_dpi, is avail, then try to
                # guess the page dpi, if that also fails, then set to min
                # or max allowed size
                if per_page_dpi is not None and dpi:
                    if verbose:
                        print('Trying document level dpi:', dpi)
                    page_width = imwidth / (dpi / 72)

                # If that didn't work, guess
                if page_width <= PDFA_MIN_UNITS or page_width >= PDFA_MAX_UNITS:
                    page_dpi = guess_dpi(imwidth, imheight,
                                         expected_format=(8.27, 11.69),
                                         round_to=(72, 96, 150, 300, 600))
                    if verbose:
                        print('Guessing DPI:', dpi)
                    page_width = imwidth / (page_dpi / 72)

                # If even guessing fails, let's just set minimal values since
                # this typically only happens for really tiny images
                if page_width <= PDFA_MIN_UNITS or page_width >= PDFA_MAX_UNITS:
                    page_width = PDFA_MIN_UNITS + 1
                    page_height = PDFA_MIN_UNITS + 1

                # Add warning/error
                if errors is not None:
                    errors.add(RECODE_RUNTIME_WARNING_INVALID_PAGE_SIZE)

            scaler = page_width / imwidth

            ppi = 72. / scaler

            width = page_width
            height = imheight * scaler

        font_scaler = 1
        if hocr_dpi is not None:
            font_scaler = hocr_dpi / ppi
        else:
            font_scaler = 72. / ppi

        word_data = hocr_page_to_word_data(hocr_page, font_scaler)
        render.AddImageHandler(word_data, width, height, ppi=ppi, hocr_ppi=hocr_dpi)

        reporting_page_count += 1


    if verbose:
        print('Finished page generation at', datetime.utcnow().isoformat())
        print('Creating text pages took %.4f seconds' % (time() - last_time))


    if reporter and reporting_page_count != 0:
        current_time = time()
        ms = int(((current_time - last_time) / reporting_page_count) * 1000)

        data = json.dumps({'text_pages': {'count': reporting_page_count,
                                              'time-per': ms}})
        subprocess.check_output(reporter, input=data.encode('utf-8'))

    render.EndDocumentHandler()

    fp = open(save_path, 'wb+')
    fp.write(render._data)
    fp.close()


def get_timing_summary(timing_data):
    sums = {}

    # We expect this to always happen per page
    image_load_c = 0

    for v in timing_data:
        key = v[0]
        val = v[1]

        if key == 'image_load':
            image_load_c += 1

        if key not in sums:
            sums[key] = 0.

        sums[key] += val

    for k in sums.keys():
        sums[k] = sums[k] / image_load_c

    for k in sums.keys():
        # For statsd, in ms
        sums[k] = int(sums[k] * 1000)

    return sums



def insert_images_mrc(to_pdf, hocr_file, from_pdf=None, image_files=None,
        dpi=None, dpi_pages=None,
        bg_compression_flags=None, fg_compression_flags=None,
        skip_pages=None, img_dir=None, jbig2=False,
        downsample=None,
        bg_downsample=None,
        fg_downsample=None,
        denoise_mask=None, reporter=None,
        hq_pages=None, hq_bg_compression_flags=None, hq_fg_compression_flags=None,
        verbose=False, debug=False, tmp_dir=None, report_every=None,
        stop_after=None, grayscale_pdf=False,
        force_1bit_output=None,
        jpeg2000_implementation=None, mrc_image_format=None, threads=None,
        errors=None):
    hocr_iter = hocr_page_iterator(hocr_file)

    skipped_pages = 0

    last_time = time()
    timing_data = []
    reporting_page_count = 0

    downsampled = False

    #for idx, page in enumerate(to_pdf):
    for idx, hocr_page in enumerate(hocr_iter):
        if skip_pages is not None and idx in skip_pages:
            skipped_pages += 1
            continue

        idx = idx - skipped_pages

        if stop_after is not None and idx >= stop_after:
            break

        picked_dpi = None

        hocr_dpi = hocr_page_get_scan_res(hocr_page)

        if dpi_pages is not None:
            picked_dpi = dpi_pages[idx]
            if picked_dpi is None:
                picked_dpi = hocr_dpi[1]

        if picked_dpi is None:
            picked_dpi = dpi

        if picked_dpi is not None:
            picked_dpi = int(picked_dpi)

        page = to_pdf[idx]

        if from_pdf is not None:
            # TODO: Support more images and their masks, if they exist (and
            # write them to the right place in the PDF)
            t = time()

            img = from_pdf[idx].get_images()[0]
            xref = img[0]
            maskxref = img[1]

            image = from_pdf.extract_image(xref)
            imgfd = io.BytesIO()
            imgfd.write(image["image"])
            image = Image.open(imgfd)
            image.load()
            imgfd.close()

            if timing_data is not None:
                timing_data.append(('image_load', time()-t))
        else:
            t = time()
            # Do not subtract skipped pages here
            imgfile = image_files[idx+skipped_pages]

            # Potentially special path
            if imgfile.endswith('.jp2') or imgfile.endswith('.jpx'):
                image = decode_jpeg2000(imgfile, reduce_=downsample,
                        impl=jpeg2000_implementation, threads=threads, debug=debug)
                if downsample:
                    downsampled = True
            else:
                image = Image.open(imgfile)
                image.load()

            if image.mode in ('RGBA', 'LA'):
                if image.mode == 'RGBA':
                    image = image.convert('RGB')
                elif image.mode == 'LA':
                    image = image.convert('L')

            if timing_data is not None:
                timing_data.append(('image_load', time()-t))

        if grayscale_pdf and image.mode not in ('L', 'LA'):
            t = time()
            image = Image.fromarray(special_gray_convert(np.array(image)))
            if timing_data is not None:
                timing_data.append(('special_gray_convert', time()-t))

        render_hq = hq_pages[idx]

        if downsample is not None and not downsampled:
            w, h = image.size
            image.thumbnail((w/downsample, h/downsample),
                            resample=Image.LANCZOS, reducing_gap=None)
            downsampled = True

        hocr_word_data = hocr_page_to_word_data(hocr_page)

        if image.mode == '1':
            ww, hh = image.size
            mask_jb2, mask_png = encode_mrc_mask(np.array(image), tmp_dir=tmp_dir,
                    jbig2=jbig2, timing_data=timing_data, debug=debug)

            t = time()

            if jbig2:
                mask_contents = open(mask_jb2, 'rb').read()
                remove(mask_jb2)
            else:
                mask_contents = open(mask_png, 'rb').read()

            # We currently always return the PNG file
            remove(mask_png)

            page.insert_image(page.rect, stream=mask_contents,
                    width=ww, height=hh, alpha=0)

            if timing_data is not None:
                timing_data.append(('page_image_insertion', time() - t))

        elif force_1bit_output == True:
            ww, hh = image.size
            mrc_gen = create_mrc_hocr_components(image, hocr_word_data,
                    dpi=picked_dpi,
                    downsample=downsample,
                    bg_downsample=None if render_hq else bg_downsample,
                    fg_downsample=None if render_hq else fg_downsample,
                    denoise_mask=denoise_mask,
                    timing_data=timing_data, errors=errors)
            np_mask = next(mrc_gen)
            np_mask = np_mask ^ np.ones(np_mask.shape, dtype=bool)
            mask_jb2, mask_png = encode_mrc_mask(np_mask, tmp_dir=tmp_dir, jbig2=jbig2,
                    timing_data=timing_data, debug=debug)

            if jbig2:
                mask_contents = open(mask_jb2, 'rb').read()
                remove(mask_jb2)
            else:
                mask_contents = open(mask_png, 'rb').read()

            # We currently always return the PNG file
            remove(mask_png)

            page.insert_image(page.rect, stream=mask_contents,
                    width=ww, height=hh, alpha=0)

            if timing_data is not None:
                timing_data.append(('page_image_insertion', time() - t))
        else:
            mrc_gen = create_mrc_hocr_components(image, hocr_word_data,
                    dpi=picked_dpi,
                    downsample=downsample,
                    bg_downsample=None if render_hq else bg_downsample,
                    fg_downsample=None if render_hq else fg_downsample,
                    denoise_mask=denoise_mask,
                    timing_data=timing_data, errors=errors)


            # TODO: keep all these files on disk, and insert them into the pager
            # later? maybe? or just saveIncr()
            # TODO: maybe call the encode_mrc_{mask,foreground,background}
            # separately from here so that we can free the arrays sooner (and even
            # get the images separately from the create_mrc_hocr_components call)

            fast_insert_image_ok = jbig2 and image.mode in ('L', 'RGB')

            mask_f, bg_f, bg_s, fg_f, fg_s = encode_mrc_images(mrc_gen,
                    bg_compression_flags=hq_bg_compression_flags if render_hq else bg_compression_flags,
                    fg_compression_flags=hq_fg_compression_flags if render_hq else fg_compression_flags,
                    tmp_dir=tmp_dir, jbig2=jbig2, timing_data=timing_data,
                    jpeg2000_implementation=jpeg2000_implementation,
                    mrc_image_format=mrc_image_format,
                    embedded_jbig2=fast_insert_image_ok,
                    threads=threads,
                    debug=debug)

            if img_dir is not None:
                shutil.copy(mask_f, join(img_dir, '%.6d_mask.jbig2' % idx))
                shutil.copy(bg_f, join(img_dir, '%.6d_bg.jp2' % idx))
                shutil.copy(fg_f, join(img_dir, '%.6d_fg.jp2' % idx))


            t = time()
            bg_contents = open(bg_f, 'rb').read()
            if not jbig2 or image.mode not in ('L', 'RGB'):
                # Tell PyMuPDF about width/height/alpha since it's faster this way
                page.insert_image(page.rect, stream=bg_contents, mask=None,
                    overlay=False, width=bg_s[0], height=bg_s[1], alpha=0)
            else:
                fast_insert_image(page, page.rect, stream=bg_contents,
                                  mask=None, width=bg_s[0], height=bg_s[1],
                                  stream_fmt=mrc_image_format,
                                  gray=image.mode == 'L')

            fg_contents = open(fg_f, 'rb').read()
            mask_contents = open(mask_f, 'rb').read()

            # Tell PyMuPDF about width/height/alpha since it's faster this way
            if not jbig2 or image.mode not in ('L', 'RGB'):
                page.insert_image(page.rect, stream=fg_contents, mask=mask_contents,
                        overlay=True, width=fg_s[0], height=fg_s[1], alpha=0)
            else:
                fast_insert_image(page, page.rect, stream=fg_contents,
                                  mask=mask_contents, width=fg_s[0], height=fg_s[1],
                                  stream_fmt=mrc_image_format,
                                  gray=image.mode == 'L')

            # Remove leftover files
            remove(mask_f)
            remove(bg_f)
            remove(fg_f)
            if timing_data is not None:
                timing_data.append(('page_image_insertion', time() - t))

        reporting_page_count += 1

        if report_every is not None and reporting_page_count % report_every == 0:
            print('Processed %d PDF pages.' % idx)
            sys.stdout.flush()

            timing_sum = get_timing_summary(timing_data)
            timing_data = []

            if reporter:
                current_time = time()
                ms = int(((current_time - last_time) / reporting_page_count) * 1000)

                data = json.dumps({'compress_pages': {'count': reporting_page_count,
                                                 'time-per': ms},
                                   'page_time_breakdown': timing_sum})
                subprocess.check_output(reporter, input=data.encode('utf-8'))

                # Reset chunk timer
                last_time = time()
                # Reset chunk counter
                reporting_page_count = 0


    if reporter and reporting_page_count != 0:
        current_time = time()
        ms = int(((current_time - last_time) / reporting_page_count) * 1000)

        timing_sum = get_timing_summary(timing_data)

        data = json.dumps({'compress_pages': {'count': reporting_page_count,
                                         'time-per': ms},
                           'page_time_breakdown': timing_sum})
        subprocess.check_output(reporter, input=data.encode('utf-8'))

    if verbose:
        summary = get_timing_summary(timing_data)
        print('MRC time breakdown:', summary)


def insert_images(from_pdf, to_pdf, mode, report_every=None, stop_after=None):
    # TODO: This hasn't been updated, should fix this up, only MRC is tested
    # really.
    # TODO: implement img_dir here

    for idx, page in enumerate(to_pdf):
        # XXX: TODO: FIXME: MEGAHACK: For some reason the _imgonly PDFs
        # generated by us have all images on all pages according to pymupdf, so
        # hack around that for now.
        img = sorted(from_pdf.getPageImageList(idx))[idx]
        #img = from_pdf.getPageImageList(idx)[0]

        xref = img[0]
        maskxref = img[1]
        if mode == IMAGE_MODE_PASSTHROUGH:
            image = from_pdf.extract_image(xref)
            page.insert_image(page.rect, stream=image["image"], overlay=False)
        elif mode == IMAGE_MODE_PIXMAP:
            pixmap = fitz.Pixmap(from_pdf, xref)
            page.insert_image(page.rect, pixmap=pixmap, overlay=False)

        if stop_after is not None and idx >= stop_after:
            break

        if report_every is not None and idx % report_every == 0:
            print('Processed %d PDF pages.' % idx)
            sys.stdout.flush()


# TODO: Document these options (like in bin/recode_pdf)
def recode(from_pdf=None, from_imagestack=None, dpi=None, hocr_file=None,
        scandata_file=None, out_pdf=None, out_dir=None,
        reporter=None,
        grayscale_pdf=False,
        force_1bit_output=False,
        image_mode=IMAGE_MODE_MRC, jbig2=False, verbose=False, debug=False,
        tmp_dir=None,
        report_every=None, stop_after=None,
        jpeg2000_implementation=JPEG2000_IMPL_PILLOW,
        bg_compression_flags=None, fg_compression_flags=None,
        mrc_image_format=None,
        downsample=None,
        bg_downsample=None,
        fg_downsample=None,
        denoise_mask=None,
        hq_pages=None,
        hq_bg_compression_flags=None, hq_fg_compression_flags=None,
        threads=None,
        render_text_lines=False,
        metadata_url=None, metadata_title=None, metadata_author=None,
        metadata_creator=None, metadata_language=None,
        metadata_subject=None, metadata_creatortool=None):
    # TODO: document that the scandata document dpi will override the dpi arg
    # TODO: Take hq-pages and reporter arg and change format (as lib call we
    # don't want to pass that as one string, I guess?)

    errors = set()

    in_pdf = None
    if from_pdf:
        in_pdf = fitz.open(from_pdf)

    image_files = None
    if from_imagestack:
        image_files = sorted(glob(from_imagestack))

    hocr_file = hocr_file
    outfile = out_pdf

    stop = stop_after
    if stop is not None:
        stop -= 1

    if verbose:
        from numpy.core._multiarray_umath import __cpu_features__ as cpu_have
        cpu = cpu_have
        for k, v in cpu.items():
            if v:
                print('\t', k)


    reporter = reporter.split(' ') if reporter else None # TODO: overriding

    start_time = time()

    scandata_doc_dpi = None

    # Figure out if we have scandata, and figure out if we want to skip pages
    # based on scandata.
    skip_pages = []
    dpi_pages = None
    if scandata_file is not None:
        skip_pages = scandata_xml_get_skip_pages(scandata_file)
        dpi_pages = scandata_xml_get_dpi_per_page(scandata_file)
        scandata_doc_dpi = scandata_xml_get_document_dpi(scandata_file)

        if scandata_doc_dpi is not None:
            # Let's prefer the DPI in the scandata file over the provided DPI
            dpi = scandata_doc_dpi

    # XXX: Maybe use a buffer, since the file is typically quite small
    fd, tess_tmp_path = mkstemp(prefix='pdfrenderer', suffix='.pdf', dir=tmp_dir)
    os.close(fd)

    if verbose:
        print('Creating text only PDF')

    # 1. Create text-only PDF from hOCR first, but honour page sizes of in_pdf
    create_tess_textonly_pdf(hocr_file, tess_tmp_path, in_pdf=in_pdf,
            image_files=image_files, dpi=dpi,
            skip_pages=skip_pages, dpi_pages=dpi_pages,
            reporter=reporter,
            verbose=verbose, debug=debug, stop_after=stop,
            render_text_lines=render_text_lines,
            tmp_dir=tmp_dir,
            jpeg2000_implementation=jpeg2000_implementation,
            errors=errors)

    if verbose:
        print('Inserting (and compressing) images')
    # 2. Load tesseract PDF and stick images in the PDF
    # We open the generated file but do not modify it in place
    outdoc = fitz.open(tess_tmp_path)

    HQ_PAGES = [False for x in range(outdoc.page_count)]
    if hq_pages is not None:
        index_range = map(int, hq_pages.split(','))
        for i in index_range:
            # We want 0-indexed, not 1-indexed, but not negative numbers we want
            # to remain 1-indexed.
            if i > 0:
                i = i - 1

            if abs(i) >= len(HQ_PAGES):
                # Page out of range, silently ignore for automation purposes.
                # We don't want scripts that call out tool to worry about how
                # many a PDF has exactly. E.g. if 1,2,3,4,-4,-3,-2,-1 is passed,
                # and a PDF has only three pages, let's just set them all to HQ
                # and not complain about 4 and -4 being out of range.
                continue

            # Mark page as HQ
            HQ_PAGES[i] = True


    if verbose:
        print('Converting with image mode:', image_mode)
    if image_mode == 2:
        insert_images_mrc(outdoc, hocr_file,
                          from_pdf=in_pdf,
                          image_files=image_files,
                          dpi=dpi,
                          dpi_pages=dpi_pages,
                          bg_compression_flags=bg_compression_flags,
                          fg_compression_flags=fg_compression_flags,
                          skip_pages=skip_pages,
                          img_dir=out_dir,
                          jbig2=jbig2,
                          downsample=downsample,
                          bg_downsample=bg_downsample,
                          fg_downsample=fg_downsample,
                          denoise_mask=denoise_mask,
                          reporter=reporter,
                          hq_pages=HQ_PAGES,
                          hq_bg_compression_flags=hq_bg_compression_flags,
                          hq_fg_compression_flags=hq_fg_compression_flags,
                          verbose=verbose,
                          debug=debug,
                          tmp_dir=tmp_dir,
                          report_every=report_every,
                          stop_after=stop,
                          grayscale_pdf=grayscale_pdf,
                          force_1bit_output=force_1bit_output,
                          jpeg2000_implementation=jpeg2000_implementation,
                          mrc_image_format=mrc_image_format,
                          threads=threads,
                          errors=errors)
    elif image_mode in (0, 1):
        # TODO: Update this codepath
        insert_images(in_pdf, outdoc, mode=image_mode,
                report_every=report_every, stop_after=stop)
    elif image_mode == 3:
        # 3 = skip
        pass

    # 3. Add PDF/A compliant data
    write_pdfa(outdoc)

    if scandata_file is not None:
        # XXX: we parse scandata twice now, let's not do that
        # 3b. Write page labels from scandata file, if present
        write_page_labels(outdoc, scandata_file, errors=errors)


    lang_if_any = metadata_language[0] if metadata_language else None
    write_basic_ua(outdoc, language=lang_if_any)

    # 4. Write metadata
    extra_metadata = {}
    if metadata_url:
        extra_metadata['url'] = metadata_url
    if metadata_title:
        extra_metadata['title'] = metadata_title
    if metadata_creator:
        extra_metadata['creator'] = metadata_creator
    if metadata_author:
        extra_metadata['author'] = metadata_author
    if metadata_language:
        extra_metadata['language'] = metadata_language
    if metadata_subject:
        extra_metadata['subject'] = metadata_subject
    if metadata_creatortool:
        extra_metadata['creatortool'] = metadata_creatortool
    write_metadata(in_pdf, outdoc, extra_metadata=extra_metadata)

    # 5. Save
    mupdf_warnings = fitz.TOOLS.mupdf_warnings()
    if mupdf_warnings:
        print('mupdf warnings:', repr(mupdf_warnings))
    if verbose:
        print('Saving PDF now')

    t = time()
    outdoc.save(outfile, deflate=True, pretty=True)
    save_time_ms = int((time() - t)*1000)
    if reporter:
        data = json.dumps({'time_to_save': {'time': save_time_ms}})
        subprocess.check_output(reporter, input=data.encode('utf-8'))

    end_time = time()
    print('Processed %d pages at %.2f seconds/page' % (len(outdoc),
        (end_time - start_time) / len(outdoc)))

    if from_pdf is not None:
        oldsize = os.path.getsize(from_pdf)
    else:
        bytesum = 0
        skipped_pages = 0
        for idx, fname in enumerate(image_files):
            if skip_pages is not None and idx in skip_pages:
                skipped_pages += 1
                continue

            if stop_after is not None and (idx - skipped_pages) > stop_after:
                break

            bytesum += os.path.getsize(fname)

        oldsize = bytesum

    newsize = os.path.getsize(out_pdf)
    compression_ratio  = oldsize / newsize
    if verbose:
        print('Compression ratio: %f' % (compression_ratio))

    # 5. Remove leftover files
    outdoc.close()
    remove(tess_tmp_path)

    return {'errors': errors,
            'compression_ratio': compression_ratio}



# archive-pdf-tools
# Copyright (C) 2020-2021, Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Merlijn Boris Wolf Wajer <merlijn@archive.org>

import numpy as np
from skimage.color import rgb2hsv

perc2val = lambda x: (x*255)/100

def level_arr(arr, minv=0, maxv=255):
    interval = (maxv/255.) - (minv/255.)
    arr_zero = arr < minv
    arr_max = arr > maxv
    arr[::] = ((arr[::] - minv) / interval)
    arr[arr_zero] = 0
    arr[arr_max] = 255
    return arr


# Straight forward port of color2Gray.sh script
# We might be able to do better, but there are only a few users of this script
# in the archive.org currently, so more time has not been invested in finding
# alternative or better ways.
def special_gray_convert(imd):
    components = ('r', 'g', 'b')

    d = {}
    for i, k in enumerate(components):
        for fun in ['min', 'max', 'mean', 'std']:
            d[k + '_' + fun] = getattr(np, fun)(imd[:,:,i]) / 255.

    bright_adjust = round(d['r_mean'] * d['g_mean'] * d['b_mean'] /
                    (d['b_max']*(1-d['r_std'])*(1-d['g_std'])*(1-d['b_std'])), 4)

    low_thres = min(int((196 * d['r_min']+14.5)/1), 50)

    high_thres = {
            'r': min(int((35.66*bright_adjust+48.5)/1), 95),
            'g': min(int((39.22*bright_adjust+44.5)/1), 95),
            'b': min(int((45.16*bright_adjust+36.5)/1), 95),
            }

    new_imd = np.copy(imd)
    for i, c in enumerate(components):
        new_imd[:,:,i] = level_arr(new_imd[:,:,i],
                                   minv=perc2val(low_thres),
                                   maxv=perc2val(high_thres[c]))

    hsv = rgb2hsv(new_imd)
    # Calculate the 'L' from 'HSL' as L = S * (1 - V/2)
    l = hsv[:,:,2] * (1 - (hsv[:,:,1]/2))
    return np.array(l * 255, dtype=np.uint8)





# Straight port for tesseract/src/api/pdfrenderer.cpp
# at commit 2d6f38eebf9a14d9fbe65d785f0d7bd898ff46cb
#
# Please look at the code in there for documentation, etc.
#
# License for this file is Apache 2.0:
#
#
#          Apache License
#    Version 2.0, January 2004
# http://www.apache.org/licenses/

import pkg_resources

from math import atan, atan2, cos, sin
import numpy as np
import zlib
import datetime
import sys

# Lives at https://git.archive.org/merlijn/archive-hocr-tools
from hocr.parse import (hocr_page_iterator, hocr_page_to_word_data,
        hocr_page_get_dimensions,
        WRITING_DIRECTION_UNSPECIFIED, WRITING_DIRECTION_LEFT_TO_RIGHT,
        WRITING_DIRECTION_RIGHT_TO_LEFT, WRITING_DIRECTION_TOP_TO_BOTTOM)


K_CHAR_WIDTH = 2
K_MAX_BYTES_PER_CODEPOINTS = 20

VERSION = '0.0.1'
SOFTWARE = 'Archive.org hOCR to PDF renderer (based on Tesseract)'

class TessPDFRenderer(object):

    def __init__(self, textonly=True, image_list=None, render_text_lines=False):
        self.textonly = textonly
        self.render_text_lines = render_text_lines

        self._obj = 0
        self._offsets = [0]

        self._pages = []

        self._data = b''

    def AppendPDFObjectDIY(self, object_size):
        self._offsets.append(object_size + self._offsets[-1])
        self._obj += 1

    def AppendPDFObject(self, data):
        self.AppendPDFObjectDIY(len(data))
        self.AppendString(data)

    def AppendString(self, s):
        self._data += s

    def AppendData(self, s):
        self._data += s

    def GetPDFTextObjects(self, word_data, width, height, ppi, hocr_ppi=None):
        # Stub values
        old_x = 0.0
        old_y = 0.0
        old_fontsize = 0
        old_writing_direction = WRITING_DIRECTION_LEFT_TO_RIGHT
        new_block = True
        fontsize = 0
        a = 1.
        b = 0.
        c = 0.
        d = 1.

        pdf_str = bytes()
        pdf_str += b'q ' + floatbytes(prec(width), prec=3) + b' 0 0 ' + floatbytes(prec(height), prec=3) + b' 0 0 cm'

        if not self.textonly:
            pdf_str += b' /Im1 Do'

        pdf_str += b' Q\n';

        line_x1 = 0
        line_y1 = 0
        line_x2 = 0
        line_y2 = 0

        for paragraph in word_data:

            # See if we want to skip this line (if it's just a space with a big
            # bounding box, we don't want to include them, as they mess up text
            # selection.
            linetext = ''
            for line in paragraph['lines']:
                for word in line['words']:
                    # XXX: We could also skip words if the word confidence is
                    # particularly low.
                    #if word['confidence'] < 75:
                    #    print('Skipping word with low confidence.')
                    #    continue
                    for char in word['text']:
                        linetext += char
            if linetext.strip() == '':
                continue


            if self.render_text_lines:
                # Use this instead of b'BT\n3 Tr' if you want to see the text
                pdf_str += b'BT\n0 Tr'
            else:
                pdf_str += b'BT\n3 Tr'
            old_fontsize = 0
            new_block = True

            for line in paragraph['lines']:
                first_word_of_line = True
                for word in line['words']:
                    if first_word_of_line:
                        bx1, by1, bx2, by2 = line['bbox']
                        slope, const = line['baseline']

                        x1 = bx1
                        y1 = by2 + const
                        x2 = bx2
                        dx = x2-x1
                        y2 = y1 + slope*(dx)

                        line_x1, line_y1, line_x2, line_y2 = \
                                ClipBaseline(ppi, x1, y1, x2, y2)

                        writing_direction = word['writing_direction']
                        if writing_direction == WRITING_DIRECTION_UNSPECIFIED:
                            writing_direction = WRITING_DIRECTION_LEFT_TO_RIGHT

                    word_x1, word_y1, word_x2, word_y2 = word['bbox']

                    x, y, word_length = GetWordBaseline(writing_direction, ppi, height,
                            word_x1, word_y1, word_x2, word_y2,
                            line_x1, line_y1,
                            line_x2, line_y2)

                    if (writing_direction != old_writing_direction) or new_block:
                        a, b, c, d = \
                                AffineMatrix(writing_direction, line_x1, line_y1, line_x2, line_y2)
                        pdf_str += (b' ' + floatbytes(prec(a)) +
                                    b' ' + floatbytes(prec(b)) +
                                    b' ' + floatbytes(prec(c)) +
                                    b' ' + floatbytes(prec(d)) +
                                    b' ' + floatbytes(prec(x)) +
                                    b' ' + floatbytes(prec(y)) +
                                    b' Tm ')

                        new_block = False
                    else:
                        dx = x - old_x
                        dy = y - old_y
                        pdf_str += b' ' + floatbytes(prec(dx * a + dy * b))
                        pdf_str += b' ' + floatbytes(prec(dx * c + dy * d))
                        pdf_str += b' Td '

                        first_word_of_line = False

                    old_x = x;
                    old_y = y;
                    old_writing_direction = writing_direction

                    fontsize = word['fontsize']

                    kDefaultFontsize = 8;
                    if fontsize <= 0:
                        line_height = abs(line_y2 - line_y1)
                        fontsize = line_height

                        if fontsize <= 0:
                            fontsize = kDefaultFontsize

                    if fontsize != old_fontsize:
                        pdf_str += b'/f-0-0 ' + str(fontsize).encode('ascii') + b' Tf ';
                        old_fontsize = fontsize;

                    pdf_word = b''
                    pdf_word_len = 0

                    for char in word['text']:
                        codepoint = ord(char)
                        ok, utf16 = CodepointToUtf16be(codepoint)
                        if ok:
                            pdf_word += utf16
                            pdf_word_len += 1

                    if True: # res_is->IsAtBeginningOf(RIL_WORD)
                        pdf_word += b'0020'
                        pdf_word_len += 1

                    if word_length > 0 and pdf_word_len > 0:
                        h_stretch = K_CHAR_WIDTH * prec(100.0 * word_length / (fontsize * pdf_word_len))
                        pdf_str += floatbytes(h_stretch) + b' Tz'
                        pdf_str += b' [ <' + pdf_word
                        pdf_str += b'> ] TJ'

                # Last word in the line
                pdf_str += b' \n';

            # Last line the block
            pdf_str += b'ET\n'


        return pdf_str

    def BeginDocumentHandler(self):
        self.AppendPDFObject(b'%PDF-1.5\n%\xDE\xAD\xBE\xEB\n');
        self.AppendPDFObject(b'1 0 obj\n'
                             b'<<\n'
                             b'  /Type /Catalog\n'
                             b'  /Pages 2 0 R\n'
                             b'>>\nendobj\n')
        self.AppendPDFObject(b'');
        self.AppendPDFObject(b'3 0 obj\n'
                             b'<<\n'
                             b'  /BaseFont /GlyphLessFont\n'
                             b'  /DescendantFonts [ 4 0 R ]\n'
                             b'  /Encoding /Identity-H\n'
                             b'  /Subtype /Type0\n'
                             b'  /ToUnicode 6 0 R\n'
                             b'  /Type /Font\n'
                             b'>>\n'
                             b'endobj\n')

        self.AppendPDFObject(b'4 0 obj\n'
                             b"<<\n"
                             b"  /BaseFont /GlyphLessFont\n"
                             b"  /CIDToGIDMap 5 0 R\n"
                             b"  /CIDSystemInfo\n"
                             b"  <<\n"
                             b"     /Ordering (Identity)\n"
                             b"     /Registry (Adobe)\n"
                             b"     /Supplement 0\n"
                             b"  >>\n"
                             b"  /FontDescriptor 7 0 R\n"
                             b"  /Subtype /CIDFontType2\n"
                             b"  /Type /Font\n"
                             b"  /DW " + str(1000 // K_CHAR_WIDTH).encode('ascii') + b'\n' +
                             b">>\n"
                             b"endobj\n")

        kCIDToGIDMapSize = 2 * (1 << 16);
        cidtogidmap = np.ndarray(kCIDToGIDMapSize, dtype='<u1')
        cidtogidmap[:] = 0
        cidtogidmap[1::2] = 1

        compressed = zlib.compress(cidtogidmap.tobytes())
        complen = len(compressed)
        stream = bytes()

        stream += b'5 0 obj\n'
        stream += b'<<\n'
        stream += b'  /Length ' + str(complen).encode('ascii') + b' /Filter /FlateDecode\n'
        stream += b'>>\n'
        stream += b'stream\n'
        self.AppendString(stream)

        objsize = len(stream)
        self.AppendString(compressed)
        objsize += complen
        endstream_obj = b'endstream\nendobj\n'
        self.AppendString(endstream_obj)
        objsize += len(endstream_obj)
        self.AppendPDFObjectDIY(objsize)

        stream2 = (b'/CIDInit /ProcSet findresource begin\n'
                  b'12 dict begin\n'
                  b'begincmap\n'
                  b'/CIDSystemInfo\n'
                  b'<<\n'
                  b'  /Registry (Adobe)\n'
                  b'  /Ordering (UCS)\n'
                  b'  /Supplement 0\n'
                  b'>> def\n'
                  b'/CMapName /Adobe-Identify-UCS def\n'
                  b'/CMapType 2 def\n'
                  b'1 begincodespacerange\n'
                  b'<0000> <FFFF>\n'
                  b'endcodespacerange\n'
                  b'1 beginbfrange\n'
                  b'<0000> <FFFF> <0000>\n'
                  b'endbfrange\n'
                  b'endcmap\n'
                  b'CMapName currentdict /CMap defineresource pop\n'
                  b'end\n'
                  b'end\n')

        stream = b'6 0 obj\n'
        stream += b'<< /Length ' + str(len(stream2)).encode('ascii') + b' >>\n'
        stream += b'stream\n' + stream2
        stream += b'endstream\n'
        stream += b'endobj\n'
        self.AppendPDFObject(stream)

        stream = (
          b'7 0 obj\n'
          b'<<\n'
          b'  /Ascent 1000\n'
          b'  /CapHeight 1000\n'
          b'  /Descent -1\n'
          b'  /Flags 5\n'
          b'  /FontBBox  [ 0 0 ' + str(1000 // K_CHAR_WIDTH).encode('ascii') + b' 1000 ]\n'
          b'  /FontFile2 8 0 R\n'
          b'  /FontName /GlyphLessFont\n'
          b'  /ItalicAngle 0\n'
          b'  /StemV 80\n'
          b'  /Type /FontDescriptor\n'
          b'>>\n'
          b'endobj\n')
        self.AppendPDFObject(stream)

        fontstream = pkg_resources.resource_string('internetarchivepdf', "data/tesseract.ttf")
        stream = (
          b'8 0 obj\n'
          b'<<\n'
          b'  /Length ' + str(len(fontstream)).encode('ascii') + b'\n'
          b'  /Length1 ' + str(len(fontstream)).encode('ascii') + b'\n'
          b'>>\n'
          b'stream\n')
        self.AppendString(stream)
        objsize  = len(stream)
        self.AppendData(fontstream)
        objsize += len(fontstream)
        self.AppendString(endstream_obj)
        objsize += len(endstream_obj)
        self.AppendPDFObjectDIY(objsize)

    def EndDocumentHandler(self, title='No title provided'):
        kPagesObjectNumber = 2

        self._offsets[kPagesObjectNumber] = self._offsets[-1]

        stream = bytes()
        stream += str(kPagesObjectNumber).encode('ascii') + b' 0 obj\n<<\n  /Type /Pages\n  /Kids [ '
        self.AppendString(stream)

        pages_objsize = len(stream)
        for i in range(len(self._pages)):
            stream = bytes()
            stream += str(self._pages[i]).encode('ascii') + b' 0 R '
            self.AppendString(stream)
            pages_objsize += len(stream)

        stream = bytes()
        stream += b']\n  /Count ' + str(len(self._pages)).encode('ascii') + b'\n>>\nendobj\n'
        self.AppendString(stream)
        pages_objsize += len(stream)
        self._offsets[-1] += pages_objsize

        utf16_title = b'FEFF'
        for c in title:
            codepoint = ord(c)
            ok, utf16 = CodepointToUtf16be(codepoint)
            if not ok:
                raise Exception('Wait what')
            if ok:
                utf16_title += utf16

        create_date = datetime.datetime.utcnow().strftime('%Y%m%d%H%M%S').encode('ascii')

        stream = bytes()
        stream += (
                str(self._obj).encode('ascii') + b' 0 obj\n'
                b'<<\n'
                b'  /Producer (' + SOFTWARE.encode('ascii') + b' ' + VERSION.encode('ascii') + b')\n'
                b'  /CreationDate (D:' + create_date + b')\n'
                b'  /Title <' + utf16_title + b'>\n'
                b'>>\n'
                b'endobj\n')
        self.AppendPDFObject(stream)
        stream = bytes()
        stream += b'xref\n0 ' + str(self._obj).encode('ascii') + b'\n0000000000 65535 f \n';
        self.AppendString(stream)
        for i in range(1, self._obj):
            stream = bytes()
            stream += ('%.10d' % self._offsets[i]).encode('ascii') + b' 00000 n \n'
            self.AppendString(stream)

        stream = bytes()
        stream += (b'trailer\n<<\n  /Size ' + str(self._obj).encode('ascii') + b'\n'
                   b'  /Root 1 0 R\n'
                   b'  /Info ' + str(self._obj - 1).encode('ascii') + b' 0 R\n'
                   b'>>\nstartxref\n' + str(self._offsets[-1]).encode('ascii') +
                   b'\n%%EOF\n')
        self.AppendString(stream)

    def AddImageHandler(self, word_data, width, height, ppi, hocr_ppi=None):
        xobject = bytes()
        stream = bytes()

        # We don't want to actually add the image itself -- one of the
        # reasons for this code to exist is to do the same _without_ adding
        # the image, but just for reference, here is the old code:
        # if !textonly
        #     xobject += b'/XObject << /Im1 ' + str((self.obj_ + 2)).encode('ascii') + b' 0 R >>\n'

        stream += (
          str(self._obj).encode('ascii') + b' 0 obj\n'
          b'<<\n'
          b'  /Type /Page\n'
          b'  /Parent 2 0 R\n'
          b'  /MediaBox [0 0 ' + str(width).encode('ascii')+ b' ' + str(height).encode('ascii') + b']\n'
          b'  /Contents ' + str(self._obj + 1).encode('ascii') + b' 0 R\n'
          b'  /Resources\n'
          b'  <<\n'
          b'    ' + xobject +
          b'    /ProcSet [ /PDF /Text /ImageB /ImageI /ImageC ]\n'
          b'    /Font << /f-0-0 3 0 R >>\n'
          b'  >>\n'
          b'>>\n'
          b'endobj\n')
        self._pages.append(self._obj)
        self.AppendPDFObject(stream)

        pdftext = self.GetPDFTextObjects(word_data, width, height, ppi, hocr_ppi=hocr_ppi)
        comp_pdftext = zlib.compress(pdftext)
        stream = bytes()

        stream += (
          str(self._obj).encode('ascii') + b' 0 obj\n'
          b'<<\n'
          b'  /Length ' + str(len(comp_pdftext)).encode('ascii') + b' /Filter /FlateDecode\n'
          b'>>\n'
          b'stream\n')
        self.AppendString(stream)
        objsize = len(stream)

        self.AppendData(comp_pdftext)
        objsize += len(comp_pdftext)

        b2 = (b"endstream\n"
             b"endobj\n")
        self.AppendString(b2)
        objsize += len(b2)
        self.AppendPDFObjectDIY(objsize)

        if False: # textonly
            pass

        return True




# Helper function to prevent from writing scientific notation to PDF file
def prec(x):
    kPrecision = 1000.0
    a = round(x * kPrecision) / kPrecision
    if a == -0:
        return 0.
    return a


def dist2(x1, y1, x2, y2):
    return (x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1)


def GetWordBaseline(writing_direction, ppi, height,
                    word_x1, word_y1, word_x2, word_y2,
                    line_x1, line_y1, line_x2, line_y2):
    if (writing_direction == WRITING_DIRECTION_RIGHT_TO_LEFT):
        tmp = word_x1
        word_x1 = word_x2
        word_x2 = tmp

        tmp = word_y1
        word_y1 = word_y2
        word_y2 = tmp

    word_length = 0.
    x = 0.
    y = 0.
    px = word_x1
    py = word_y1
    l2 = float(dist2(line_x1, line_y1, line_x2, line_y2))
    if (l2 == 0):
        x = line_x1
        y = line_y1
    else:
          t = ((px - line_x2) * (line_x2 - line_x1) +
                      (py - line_y2) * (line_y2 - line_y1)) / l2
          x = line_x2 + t * (line_x2 - line_x1)
          y = line_y2 + t * (line_y2 - line_y1)

    word_length = float(dist2(word_x1, word_y1, word_x2, word_y2) ** 0.5)
    word_length = word_length * 72.0 / ppi
    x = x * 72 / ppi
    y = height - (y * 72.0 / ppi)

    return x, y, word_length

def AffineMatrix(writing_direction, line_x1, line_y1, line_x2, line_y2):
    theta = atan2(float(line_y1 - line_y2), float(line_x2 - line_x1))
    a = cos(theta)
    b = sin(theta)
    c = -sin(theta)
    d = cos(theta)

    if writing_direction == WRITING_DIRECTION_LEFT_TO_RIGHT:
        pass
    elif writing_direction == WRITING_DIRECTION_RIGHT_TO_LEFT:
        a = -a;
        b = -b;
    elif writing_direction == WRITING_DIRECTION_TOP_TO_BOTTOM:
        # From original Tess code: TODO(jbreiden) Consider using the vertical PDF writing mode.
        pass
    else:
        raise Exception('Unknown writing direction: %d' % writing_direction)

    return a, b, c, d


def ClipBaseline(ppi, x1, y1, x2, y2):
    line_x1 = x1
    line_y1 = y1
    line_x2 = x2
    line_y2 = y2
    rise = abs(y2 - y1) * 72
    run = abs(x2 - x1) * 72
    if (rise < 2 * ppi) and (2 * ppi < run):
        line_y1 = line_y2 = (y1 + y2) / 2

    return line_x1, line_y1, line_x2, line_y2


def CodepointToUtf16be(code):
    res = None

    if (((code > 0xD7FF) and (code < 0xE000)) or (code > 0x10FFFF)):
        print("Dropping invalid codepoint %d\n" % code, file=sys.stderr);
        return False, res

    if (code < 0x10000):
        res = '%04X' % code
    else:
        a = code - 0x010000
        high_surrogate = (0x03FF & (a >> 10)) + 0xD800
        low_surrogate = (0x03FF & a) + 0xDC00
        res = '%04X04X' % (high_surrogate, low_surrogate)

    return True, res.encode('ascii')


def floatbytes(v, prec=8):
    fmt_str = '{:.%df}' % prec
    return fmt_str.format(v).encode('ascii')



# Code inspired / ported from:
#
# 1. https://arxiv.org/pdf/1905.13038.pdf
# 2. https://github.com/chungkwong/binarizer
#
# The Sauvola specific formula is repeated four times in the code because I
# didn't get function calls to work fast in Cython (40s for a 15000x15000 images instead of 4s)
#
# Author: Merlijn Wajer <merlijn@archive.org>
# License: AGPL-v3

import numpy as np
cimport numpy as np
cimport cython

INTDTYPE = np.int
UINT8DTYPE = np.uint8

ctypedef np.int_t INTDTYPE_t
ctypedef np.uint8_t UINT8DTYPE_t

# Speed up the code
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
# Let's make sure all variables have a c type, otherwise performs goes out of
# the window
@cython.warn.undeclared(True)
def binarise_sauvola(np.ndarray[UINT8DTYPE_t, ndim=1] in_arr, np.ndarray[UINT8DTYPE_t, ndim=1] out_arr, int width, int height, int window_width, int window_height, double k, double R):
    """
    Perform fast Sauvola binarisation on the given input array/image.

    Args:

    * in_arr (numpy.ndarray[numpy.uint8, ndim=1]): input array, flattened
    * our_arr (numpy.ndarray[numpy.uint8, ndim=1]): output array, flattened, must be allocated already
    * width (int): width of mask
    * height int): height of mask
    * window_width (int): Sauvola window width
    * window_height(int): Sauvola window height
    * k (double): k parameter
    * R (double): R parameter

    Example usage:

    >>> h, w = img.shape
    >>> out_img = np.ndarray(img.shape, dtype=np.bool)
    >>> out_img = np.reshape(out_img, w*h)
    >>> in_img = np.reshape(img, w*h)
    >>> binarise_sauvola(in_img, out_img, w, h, window_size, window_size, k, R)
    >>> out_img = np.reshape(out_img, (h, w))
    >>> out_img = np.invert(out_img)
    """
    cdef int i, l, r, o, u, j, dr0, dr1, dr2, dr3
    cdef int index_top, index_bottom, win_top, win_bottom, index, bottom_max
    cdef int top, bottom, win_height, sum_, count, win_right, win_left, imax
    cdef long square_sum

    cdef char formres;
    cdef double mean, variance, tmp

    cdef double k2=k*k/R/R;

    cdef np.ndarray[INTDTYPE_t, ndim=1] integral = np.zeros([width], dtype=INTDTYPE)
    cdef np.ndarray[INTDTYPE_t, ndim=1] integral_square = np.zeros([width], dtype=INTDTYPE)

    cdef INTDTYPE_t pixel


    for i in range(0, width):
        integral[i] = 0
        integral_square[i] = 0

    l = (window_width + 1) / 2
    r = window_width / 2;
    o = (window_height + 1) / 2
    u = window_height / 2

    index = 0
    imax = min(height, u)
    for i in range(0, imax):
        for j in range(0, width):
            pixel = in_arr[index]

            integral[j] += pixel
            integral_square[j] += pixel * pixel

            index += 1

    dr0 = min(window_width, width)
    dr1 = min(r, width)
    dr2 = width - l
    dr3 = min(dr2, 0)
    index_top = 0
    index_bottom = u * width

    win_top = -o
    win_bottom = u
    index = 0
    bottom_max = height + u
    while win_bottom < bottom_max:
        if win_top>=0:
            for j in range(0, width):
                pixel = in_arr[index_top]

                integral[j] -= pixel
                integral_square[j] -= pixel * pixel

                index_top += 1

            top = win_top
        else:
            top = -1

        if win_bottom < height:
            for j in range(0, width):
                pixel = in_arr[index_bottom]

                integral[j] += pixel
                integral_square[j] += pixel*pixel

                index_bottom += 1

            bottom = win_bottom
        else:
            bottom = height - 1;

        win_height = bottom-top
        sum_ = 0
        square_sum = 0
        for j in range(0, dr1):
            sum_ += integral[j];
            square_sum += integral_square[j];

        count = dr1 * win_height;
        win_right = r;
        while win_right < dr0:
            count += win_height;
            sum_ += integral[win_right];
            square_sum += integral_square[win_right];

            pixel = in_arr[index]
            if k >= 0:
                mean = sum_ / count
                variance = square_sum / count - (mean * mean)
                tmp = (pixel + (mean * (k- 1)))
                formres = ((tmp <= 0) or (tmp * tmp <= mean * mean * k2 * variance))
            else:
                mean = sum_ / count
                variance = square_sum / count - (mean*mean)
                tmp = (pixel + mean * (k - 1))
                formres = (tmp <= 0 and tmp * tmp >= mean * mean * k2 * variance)
            out_arr[index] = 0 if formres else 1

            win_right += 1
            index += 1

        win_left = win_right - window_width;
        if win_right >= width:
            while win_left < dr3:
                pixel = in_arr[index]
                if k >= 0:
                    mean = sum_ / count
                    variance = square_sum / count - (mean * mean)
                    tmp = (pixel + (mean * (k- 1)))
                    formres = ((tmp <= 0) or (tmp * tmp <= mean * mean * k2 * variance))
                else:
                    mean = sum_ / count
                    variance = square_sum / count - (mean*mean)
                    tmp = (pixel + mean * (k - 1))
                    formres = (tmp <= 0 and tmp * tmp >= mean * mean * k2 * variance)
                out_arr[index] = 0 if formres else 1

                win_left += 1
                index += 1
        else:
            while win_right < width:
                sum_ += integral[win_right] - integral[win_left];
                square_sum += integral_square[win_right] - integral_square[win_left];

                pixel = in_arr[index]
                if k >= 0:
                    mean = sum_ / count
                    variance = square_sum / count - (mean * mean)
                    tmp = (pixel + (mean * (k- 1)))
                    formres = ((tmp <= 0) or (tmp * tmp <= mean * mean * k2 * variance))
                else:
                    mean = sum_ / count
                    variance = square_sum / count - (mean*mean)
                    tmp = (pixel + mean * (k - 1))
                    formres = (tmp <= 0 and tmp * tmp >= mean * mean * k2 * variance)
                out_arr[index] = 0 if formres else 1

                win_left += 1
                win_right += 1
                index += 1

        while win_left < dr2:
            count -= win_height;
            sum_ -= integral[win_left];
            square_sum -= integral_square[win_left];

            pixel = in_arr[index]
            if k >= 0:
                mean = sum_ / count
                variance = square_sum / count - (mean * mean)
                tmp = (pixel + (mean * (k- 1)))
                formres = ((tmp <= 0) or (tmp * tmp <= mean * mean * k2 * variance))
            else:
                mean = sum_ / count
                variance = square_sum / count - (mean*mean)
                tmp = (pixel + mean * (k - 1))
                formres = (tmp <= 0 and tmp * tmp >= mean * mean * k2 * variance)
            out_arr[index] = 0 if formres else 1

            win_left += 1
            index += 1

        win_top += 1
        win_bottom += 1

    return 0



# archive-pdf-tools
# Copyright (C) 2020-2021, Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Merlijn Boris Wolf Wajer <merlijn@archive.org>

import xmltodict


def scandata_xml_get_skip_pages(xml_file):
    scandata = xmltodict.parse(open(xml_file, 'rb'))

    skip = []

    for idx in range(len(scandata['book']['pageData']['page'])):
        try:
            add_to_access_format = scandata['book']['pageData']['page'][idx]['addToAccessFormats']
            if add_to_access_format == 'false':
                skip.append(idx)
        except KeyError:
            pass

    return skip


def scandata_xml_get_page_numbers(xml_file):
    scandata = xmltodict.parse(open(xml_file, 'rb'))

    res = []

    pages = scandata['book']['pageData']['page']

    # If there is just one page, pages is not a list.
    if not isinstance(pages, list):
        pages = [pages]
    for idx in range(len(pages)):
        try:
            add_to_access_format = pages[idx]['addToAccessFormats']
            if add_to_access_format == 'false':
                continue
        except KeyError:
            pass

        pno = pages[idx].get('pageNumber', None)
        res.append(pno)

    return res


def scandata_xml_get_dpi_per_page(xml_file):
    scandata = xmltodict.parse(open(xml_file, 'rb'))

    res = []

    pages = scandata['book']['pageData']['page']

    # If there is just one page, pages is not a list.
    if not isinstance(pages, list):
        pages = [pages]
    for idx in range(len(pages)):
        try:
            add_to_access_format = pages[idx]['addToAccessFormats']
            if add_to_access_format == 'false':
                continue
        except KeyError:
            pass

        ppi = pages[idx].get('ppi', None)
        res.append(ppi)

    return res


def scandata_xml_get_document_dpi(xml_file):
    scandata = xmltodict.parse(open(xml_file, 'rb'))

    doc_ppi = scandata['book']['bookData'].get('dpi', None)

    if doc_ppi is not None:
        try:
            doc_ppi = int(doc_ppi)
        except (ValueError):
            return None

    return doc_ppi



# archive-pdf-tools
# Copyright (C) 2020-2021, Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Merlijn Boris Wolf Wajer <merlijn@archive.org>
#
# Helper functions for efficient JPEG2000 handing using Pillow and/or external
# binaries.

import sys
from os import close, remove
from subprocess import check_call, DEVNULL
from tempfile import mkstemp
from ast import literal_eval

from PIL import Image
from PIL import Jpeg2KImagePlugin

from internetarchivepdf.const import RECODE_RUNTIME_WARNING_INVALID_JP2_HEADERS
from internetarchivepdf.const import (JPEG2000_IMPL_KAKADU,
        JPEG2000_IMPL_OPENJPEG, JPEG2000_IMPL_GROK,
        JPEG2000_IMPL_PILLOW, JPEG2000_IMPLS)


KDU_COMPRESS = 'kdu_compress'
KDU_EXPAND = 'kdu_expand'
OPJ_COMPRESS = 'opj_compress'
OPJ_DECOMPRESS = 'opj_decompress'
GRK_COMPRESS = 'grk_compress'
GRK_DECOMPRESS = 'grk_decompress'

def encode_jpeg2000(image, outpath, impl, flags, tmp_dir=None, imgtype=None,
        threads=None, debug=False):
    """ Encode PIL image to JPEG2000 file

    Args:

    * image (PIL.Image): image to compress
    * impl (str): JPEG2000 implementation
    * flags list of str: encoding flags
    * threads (int): How many threads to use
    """
    if impl not in JPEG2000_IMPLS:
        raise Exception('Error: invalid jpeg2000 implementation?')

    if impl == JPEG2000_IMPL_PILLOW:
        kwargs = _jpeg2000_pillow_str_to_kwargs(flags[0])
        image.save(outpath, **kwargs)
    elif impl in (JPEG2000_IMPL_KAKADU, JPEG2000_IMPL_GROK,
            JPEG2000_IMPL_OPENJPEG):
        if impl in (JPEG2000_IMPL_KAKADU, JPEG2000_IMPL_GROK):
            fd, img_tiff = mkstemp(prefix=imgtype, suffix='.tif', dir=tmp_dir)
        else:
            fd, img_tiff = mkstemp(prefix=imgtype, suffix='.pnm', dir=tmp_dir)
        close(fd)

        image.save(img_tiff)

        args = ['-i', img_tiff, '-o', outpath]
        args += flags

        # Specify threads at the end, because some older versions of OpenJPEG do
        # not support threads, but if we specify the flags at the end, it will
        # just ignore the threads and not the variables passed before that (see
        # https://github.com/internetarchive/archive-pdf-tools/issues/41)
        args = add_impl_args(args, impl, encode=True, threads=threads)

        if debug:
            print('check_call: %s' % args, file=sys.stderr)
        check_call(args, stdout=DEVNULL, stderr=DEVNULL)

        remove(img_tiff)


def decode_jpeg2000(infile, reduce_=None, impl=JPEG2000_IMPL_PILLOW,
        tmp_dir=None, threads=None, debug=False):
    """ Decode JPEG2000 file to PIL image

    Args:

    * infile (str): Path of image to load
    * reduce_ (bool or None): Optional, reduction factor
    * impl (str): JPEG2000 implementation
    * tmp_dir (str): Temporary directory to use, if any
    * threads (int): How many theads to use

    Returns: loaded image (PIL.Image)
    """
    if impl not in JPEG2000_IMPLS:
        raise Exception('Error: invalid jpeg2000 implementation?')

    if reduce_ is not None:
        # TODO: Check if reduce_ is an int? (not a float, etc)
        reduce_ = int(reduce_ - 1)
        if reduce_ == 1:
            # Don't reduce when downsample = 1, it means we don't want to change
            # it, and it also complicates the arg handling below
            reduce_ = None

    img = None
    img_tiff = None

    if impl == JPEG2000_IMPL_PILLOW:
        img = Image.open(infile)
        if reduce_ is not None:
            img = img.reduce(reduce_)
    else:
        fd, img_tiff = mkstemp(suffix='.tif', dir=tmp_dir)
        close(fd)

        args = ['-i', infile, '-o', img_tiff]
        if reduce_ is not None:
            if impl in (JPEG2000_IMPL_KAKADU,):
                args += ['-reduce', str(reduce_ - 1)]
            if impl in (JPEG2000_IMPL_OPENJPEG, JPEG2000_IMPL_GROK):
                args += ['-r', str(reduce_ - 1)]

        args = add_impl_args(args, impl, encode=False, threads=threads)

        if debug:
            print('check_call: %s' % args, file=sys.stderr)
        check_call(args, stdout=DEVNULL, stderr=DEVNULL)

        img = Image.open(img_tiff)

    img.load()

    if sys.platform == 'win32':
        new_img = img.copy()
        img.close()
        img = new_img

    if img_tiff is not None:
        remove(img_tiff)

    return img


def get_jpeg2000_info(infile, impl, errors=None):
    size = None
    mode = None

    # Pillow reads the entire file for JPEG2000 images - just to get
    # the image size - so let's use some internal functions to speed that up.
    fd = open(infile, 'rb')
    try:
        size, mode, mimetype, dpi = Jpeg2KImagePlugin._parse_jp2_header(fd)
    except Exception:
        # JP2 lacks some info and PIL doesn't like it (Image.open
        # will not work, so use kdu_expand to create a tiff)
        if errors is not None:
            errors.add(RECODE_RUNTIME_WARNING_INVALID_JP2_HEADERS)

        img = decode_jpeg2000(infile, impl=impl)
        size = img.size
        mode = img.mode
        img = None
    finally:
        fd.close()

    return size, mode


def add_impl_args(args, impl, encode=False, threads=None):
    threads = str(threads) if threads else '1'

    # Use just one core
    if impl in (JPEG2000_IMPL_KAKADU,):
        # From kdu_expand/kdu_compress:
        # The special value of 0 may be used to specify that you want to use the
        # conventional single-threaded processing machinery -- i.e., you don't
        # want to create or use a threading environment.
        if threads == '1':
            threads = '0'
        args += ['-num_threads', threads]
        if encode:
            args = [KDU_COMPRESS] + args
        else:
            args = [KDU_EXPAND] + args
    if impl in (JPEG2000_IMPL_OPENJPEG,):
        args += ['-threads', threads]
        if encode:
            args = [OPJ_COMPRESS] + args
        else:
            args = [OPJ_DECOMPRESS] + args
    if impl in (JPEG2000_IMPL_GROK,):
        args += ['-H', threads]
        if encode:
            args = [GRK_COMPRESS] + args
        else:
            args = [GRK_DECOMPRESS] + args

    return args

def _jpeg2000_pillow_str_to_kwargs(s):
    kwargs = {}
    for en in s.split(';'):
        k, v = en.split(':', maxsplit=1)
        kwargs[k] = literal_eval(v)

    return kwargs



import numpy
from setuptools import setup, convert_path
from Cython.Build import cythonize

# Work around bugs in passing CFLAGS to the Cython compilation process
import os
os.environ['CFLAGS'] = '-Ofast -DNPY_NO_DEPRECATED_API'

if __name__ == '__main__':

    main_ns = {}
    ver_path = convert_path('internetarchivepdf/const.py')
    with open(ver_path) as ver_file:
        exec(ver_file.read(), main_ns)

    version = main_ns['__version__']
    setup(
        version = version,
        download_url='https://github.com/internetarchive/archive-pdf-tools/archive/%s.tar.gz' % version,
        include_dirs = [numpy.get_include()],
        scripts=['bin/recode_pdf', 'bin/pdf-metadata-json',
                 'bin/compress-pdf-images', 'bin/pdfcomp',
                 'bin/epub-to-pdf'],
        ext_modules = cythonize(
            ['cython/sauvola.pyx', 'cython/optimiser.pyx'],
            compiler_directives = {'language_level' : '3'},
        ),
    )



# Author: Merlijn Wajer <merlijn@archive.org>
# License: AGPL-v3

import numpy as np
cimport numpy as np
cimport cython

INTDTYPE = np.int
UINT8DTYPE = np.uint8
ctypedef np.int_t INTDTYPE_t
ctypedef np.uint8_t UINT8DTYPE_t

# Speed up the code and let's make sure all variables have a c type, otherwise performs goes out of
# the window
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.warn.undeclared(True)
def optimise_gray(np.ndarray[UINT8DTYPE_t, ndim=2] mask,
        np.ndarray[UINT8DTYPE_t, ndim=2] img, int width, int height, int n_size):
    cdef np.ndarray[UINT8DTYPE_t, ndim=2] new_img
    cdef int x, y
    cdef int val_count, val, ys, ye, xs, xe, xx, yy

    # TODO: weights for distance?

    new_img = np.copy(img)
    for y in range(0, height):
        for x in range(0, width):
            if not mask[y, x]:
                val_count = 0
                val = 0

                ys = max(0, y - n_size)
                ye = min(height, y + n_size)
                xs = max(0, x - n_size)
                xe = min(width, x + n_size)

                for yy in range(ys, ye):
                    for xx in range(xs, xe):
                        if mask[yy, xx]:
                            val += img[yy, xx]
                            val_count += 1

                for yy in range(ys, y):
                    for xx in range(xs, x):
                        val += new_img[yy, xx]
                        val_count += 1

                if val_count > 0:
                    new_img[y, x] = val / val_count
                else:
                    new_img[y, x] = 0

    return new_img


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.warn.undeclared(True)
def optimise_rgb(np.ndarray[UINT8DTYPE_t, ndim=2] mask,
        np.ndarray[UINT8DTYPE_t, ndim=3] img, int width, int height, int n_size):
    cdef np.ndarray[UINT8DTYPE_t, ndim=3] new_img
    cdef int x, y
    cdef int val_count, ys, ye, xs, xe, xx, yy
    cdef int r, g, b

    # TODO: weights for distance?

    new_img = np.copy(img)
    for y in range(0, height):
        for x in range(0, width):
            if not mask[y, x]:
                val_count = 0
                r = 0
                g = 0
                b = 0

                ys = max(0, y - n_size)
                ye = min(height, y + n_size)
                xs = max(0, x - n_size)
                xe = min(width, x + n_size)

                for yy in range(ys, ye):
                    for xx in range(xs, xe):
                        if mask[yy, xx]:
                            b += img[yy, xx, 0]
                            g += img[yy, xx, 1]
                            r += img[yy, xx, 2]
                            val_count += 1

                for yy in range(ys, y):
                    for xx in range(xs, x):
                        b += new_img[yy, xx, 0]
                        g += new_img[yy, xx, 1]
                        r += new_img[yy, xx, 2]
                        val_count += 1

                if val_count > 0:
                    new_img[y, x, 0] = b / val_count
                    new_img[y, x, 1] = g / val_count
                    new_img[y, x, 2] = r / val_count
                else:
                    new_img[y, x, 0] = 0
                    new_img[y, x, 1] = 0
                    new_img[y, x, 2] = 0

    return new_img

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.warn.undeclared(True)
def optimise_gray2(np.ndarray[UINT8DTYPE_t, ndim=2] mask,
        np.ndarray[UINT8DTYPE_t, ndim=2] img, int width, int height, int n_size):
    cdef np.ndarray[UINT8DTYPE_t, ndim=2] new_img
    cdef int x, y
    cdef int val_count, val, ys, ye, xs, xe, xx, yy
    cdef int ifysc, ifyec, iiysc, iiyec, ifxsc, ifxec, iixsc, iixec
    cdef int inc_fir_px_val, inc_fir_px_mask
    cdef int inc_iir_px_val

    # TODO: weights for distance?
    # This function computes a FIR and IIR version of the box blur filter incrementally
    # As seen above
    cdef np.ndarray[INTDTYPE_t, ndim=1] inc_fir_val = np.zeros([width], dtype=INTDTYPE)
    cdef np.ndarray[INTDTYPE_t, ndim=1] inc_fir_mask = np.zeros([width], dtype=INTDTYPE)
    cdef np.ndarray[INTDTYPE_t, ndim=1] inc_iir_val = np.zeros([width], dtype=INTDTYPE)

    # incremental cursors that track y-dimension FIR filter window borders
    ifysc = 0
    ifyec = 0

    # incremental cursors that track y-dimension IIR filter window borders
    iiysc = 0
    iiyec = 0

    # FIXME
    #new_img = np.empty((img.shape[0], img.shape[1]), dtype=img.dtype)
    new_img = np.copy(img)
    for y in range(0, height):
        #print('loop y')
        ys = max(0, y - n_size)
        ye = min(height, y + n_size)
        # Update y-dimension FIR window
        while ifysc < ys:
            #print('dec y')
            for x in range(0, width):
                if mask[ifysc, x]:
                    inc_fir_val[x] -= img[ifysc, x]
                    inc_fir_mask[x] -= 1
            ifysc += 1
        while ifyec < ye:
            #print('inc y')
            for x in range(0, width):
                if mask[ifyec, x]:
                    inc_fir_val[x] += img[ifyec, x]
                    inc_fir_mask[x] += 1
            ifyec += 1
        while iiysc < ys:
            for x in range(0, width):
                inc_iir_val[x] -= new_img[iiysc, x]
            iiysc += 1
        while iiyec < y:
            for x in range(0, width):
                inc_iir_val[x] += new_img[iiyec, x]
            iiyec += 1

        # incremental cursors that track x-dimension FIR filter window borders
        ifxsc = 0
        ifxec = 0

        # incremental cursors that track x-dimension IIR filter window borders
        iixsc = 0
        iixec = 0

        # incremental FIR value/mask
        inc_fir_px_val = 0
        inc_fir_px_mask = 0

        # incremental IIR value
        inc_iir_px_val = 0

        for x in range(0, width):
            #print('loop x')
            xs = max(0, x - n_size)
            xe = min(width, x + n_size)

            # Update x-dimension FIR window
            while ifxsc < xs:
                #print('dec x')
                inc_fir_px_val -= inc_fir_val[ifxsc]
                inc_fir_px_mask -= inc_fir_mask[ifxsc]
                ifxsc += 1
            while ifxec < xe:
                #print('inc x')
                inc_fir_px_val += inc_fir_val[ifxec]
                inc_fir_px_mask += inc_fir_mask[ifxec]
                ifxec += 1
            while iixsc < xs:
                inc_iir_px_val -= inc_iir_val[iixsc]
                iixsc += 1
            while iixec < x:
                inc_iir_px_val += inc_iir_val[iixec]
                iixec += 1

            #print('val', inc_fir_px_val)
            if not mask[y, x]:
                val_count = 0
                val = 0

                # Selective FIR component
                ##for yy in range(ys, ye):
                ##    for xx in range(xs, xe):
                ##        if mask[yy, xx]:
                ##            val += img[yy, xx]
                ##            val_count += 1

                ###print('prog', val, inc_fir_px_val)
                ##if val != inc_fir_px_val:
                ##    print(y, x)
                ##    print(ifysc, ifyec, ifxsc, ifxec)
                ##    print(ys, ye, xs, xe)
                ##    print(val, inc_fir_px_val)
                ##    print(val_count, inc_fir_px_mask)
                ##    print(img[ys:ye, xs:xe])
                ##    print(mask[ys:ye, xs:xe])
                ##    print(inc_fir_val[xs:xe])
                ##assert val == inc_fir_px_val
                ##assert val_count == inc_fir_px_mask
                val = inc_fir_px_val
                val_count = inc_fir_px_mask

                iir_window_size = (y - ys) * (x - xs)

                # IIR box blur over output image
                #for yy in range(ys, y):
                #    for xx in range(xs, x):
                #        val += new_img[yy, xx]
                #        val_count += 1
                #if val != inc_fir_px_val + inc_iir_px_val:
                #    print(y, x)
                #    print(iiysc, iiyec, iixsc, iixec)
                #    print(ys, y, xs, x)
                #    print(val, inc_fir_px_val + inc_iir_px_val)
                #    print(val_count, inc_fir_px_mask + iir_window_size)
                #assert val == inc_fir_px_val + inc_iir_px_val
                #assert val_count == inc_fir_px_mask + iir_window_size
                val = inc_fir_px_val + inc_iir_px_val
                val_count = inc_fir_px_mask + iir_window_size

                if val_count > 0:
                    new_img[y, x] = val / val_count
                else:
                    new_img[y, x] = 0
            #else:
            #    new_img[y, x] = img[y, x]

    return new_img



# archive-pdf-tools
# Copyright (C) 2020-2021, Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Merlijn Boris Wolf Wajer <merlijn@archive.org>
#
# Functions to (more quickly) create and modify PDFs, as well as some other low
# level PDF writing code.
#
# Some of this code is contributed by Jorj X. McKie <jorj.x.mckie@outlook.de>
# License in this file if AGPL-3 (like most of the project)
#
# For fast_insert_image, see this for more background:
# https://github.com/pymupdf/PyMuPDF/issues/1408

import pkg_resources
from math import ceil
from datetime import datetime
from xml.sax.saxutils import escape as xmlescape

from internetarchivepdf.const import COMPRESSOR_JPEG, COMPRESSOR_JPEG2000, \
        COMPRESSOR_JBIG2, PRODUCER, RECODE_RUNTIME_WARNING_INVALID_PAGE_NUMBERS
from internetarchivepdf.pagenumbers import parse_series, series_to_pdf
from internetarchivepdf.scandata import scandata_xml_get_page_numbers


JPX_TEMPL = """<<
  /Type /XObject
  /Subtype /Image
  /BitsPerComponent 8
  /Width &width
  /Height &height
  /ColorSpace /&colourspace
  /Length &len
>>"""

JPEG_TEMPL = """<<
  /Type /XObject
  /Subtype /Image
  /BitsPerComponent 8
  /Width &width
  /Height &height
  /ColorSpace /&colourspace
  /Length &len
>>"""

JBIG2_TEMPL = """<<
  /Type /XObject
  /Subtype /Image
  /BitsPerComponent 1
  /Width &width
  /Height &height
  /ColorSpace /DeviceGray
  /Length &len
>>"""


def jpx_string(stream=None, width=0, height=0, gray=True):
    if any((stream == None, width == 0, height == 0)):
        raise ValueError("invalid args")
    jpx = (
        JPX_TEMPL.replace("&width", str(width))
        .replace("&height", str(height))
        .replace("&colourspace", 'DeviceGray' if gray else 'DeviceRGB')
        .replace("&len", str(len(stream)))
    )
    return jpx


def jpg_string(stream=None, width=0, height=0, gray=True):
    if any((stream == None, width == 0, height == 0)):
        raise ValueError("invalid args")
    jpg = (
        JPX_TEMPL.replace("&width", str(width))
        .replace("&height", str(height))
        .replace("&colourspace", 'DeviceGray' if gray else 'DeviceRGB')
        .replace("&len", str(len(stream)))
    )
    return jpg


def jbig2_string(stream=None, width=0, height=0):
    if any((stream == None, width == 0, height == 0)):
        raise ValueError("invalid args")
    jbig2 = (
        JBIG2_TEMPL.replace("&width", str(width))
        .replace("&height", str(height))
        .replace("&len", str(len(stream)))
    )
    return jbig2


def fast_insert_image(page, rect=None, width=0, height=0, stream=None,
                      mask=None, stream_fmt=COMPRESSOR_JPEG2000,
                      mask_fmt=COMPRESSOR_JBIG2, gray=True):
    """Fast image insertion

    Args:

    * page: output fitz.Page
    * rect: rectangle to use
    * width: image width
    * height: image height
    * stream: image stream
    * mask: mask image stream (if any)
    * stream_fmt: COMPRESSOR_JPEG2000 or COMPRESSOR_JPEG
    * mask_fmt: COMPRESSOR_JBIG2 or None
    * gray: if the image is grayscale (otherwise RGB is assumed)
    """
    # We encode jbig2 ourselves using jbig2enc, we can't do that for ccitt
    # currently, so we rely on mupdf to do it for us, so let's not support that
    # in this code path now
    if mask_fmt not in (COMPRESSOR_JBIG2,):
        raise ValueError('mask_fmt can only be jbig2')

    # We can't handle other formats (yet)
    if stream_fmt not in (COMPRESSOR_JPEG, COMPRESSOR_JPEG2000):
        raise ValueError('stream_fmt can only be jpeg or jpeg2000')

    doc = page.parent
    nxref = doc.get_new_xref()  # make image xref in output page
    xref_stream = stream
    mask_stream = mask

    # Make object string for target page
    if stream_fmt == COMPRESSOR_JPEG2000:
        jpx_obj = jpx_string(stream=xref_stream, width=width, height=height,
                             gray=gray)
    elif stream_fmt == COMPRESSOR_JPEG:
        jpx_obj = jpg_string(stream=xref_stream, width=width, height=height,
                             gray=gray)

    doc.update_object(nxref, jpx_obj)  # give it the object definition

    # give it the image stream - unchanged compression
    doc.update_stream(nxref, stream=xref_stream, new=True, compress=False)

    # adjust image definition with correct compression info
    # this must happen AFTER stream insertion!
    if stream_fmt == COMPRESSOR_JPEG2000:
        doc.xref_set_key(nxref, "Filter", "/JPXDecode")
    elif stream_fmt == COMPRESSOR_JPEG:
        doc.xref_set_key(nxref, "Filter", "/DCTDecode")

    # if input image had a mask, we need further adjustments ...
    if mask_stream:
        nmask = doc.get_new_xref()  # need another xref in target doc

        # make smask object definition
        mask_obj = jbig2_string(stream=mask_stream, width=width, height=height)
        # and put it in mask object xref
        doc.update_object(nmask, mask_obj)

        # now insert raw mask image stream
        doc.update_stream(nmask, stream=mask_stream, new=True, compress=False)

        # and also adjust the compression filer ... AFTER stream insertion
        doc.xref_set_key(nmask, "Filter", "/JBIG2Decode")

        # we also need to tell the main image that it has a mask:
        doc.xref_set_key(nxref, "SMask", "%i 0 R" % nmask)

    # now we are ready to insert the image
    return page.insert_image(rect, xref=nxref)


# XXX: tmp.icc - pick proper one and ship it with the tool, or embed it
def write_pdfa(to_pdf):
    srgbxref = to_pdf.get_new_xref()
    to_pdf.update_object(srgbxref, """
<<
      /Alternate /DeviceRGB
      /N 3
>>
""")
    icc = pkg_resources.resource_string('internetarchivepdf', "data/tmp.icc")
    to_pdf.update_stream(srgbxref, icc, new=True)

    intentxref = to_pdf.get_new_xref()
    to_pdf.update_object(intentxref, """
<<
  /Type /OutputIntent
  /S /GTS_PDFA1
  /OutputConditionIdentifier (Custom)
  /Info (sRGB IEC61966-2.1)
  /DestOutputProfile %d 0 R
>>
""" % srgbxref)

    catalogxref = to_pdf.pdf_catalog()
    s = to_pdf.xref_object(to_pdf.pdf_catalog())
    s = s[:-2]
    s += '  /OutputIntents [ %d 0 R ]' % intentxref
    s += '>>'
    to_pdf.update_object(catalogxref, s)


def write_page_labels(to_pdf, scandata, errors=None):
    page_numbers = scandata_xml_get_page_numbers(scandata)
    res, all_ok = parse_series(page_numbers)

    # Add warning/error
    if errors is not None and not all_ok:
        errors.add(RECODE_RUNTIME_WARNING_INVALID_PAGE_NUMBERS)

    catalogxref = to_pdf.pdf_catalog()
    s = to_pdf.xref_object(to_pdf.pdf_catalog())
    s = s[:-2]
    s += series_to_pdf(res)
    s += '>>'
    to_pdf.update_object(catalogxref, s)



def write_basic_ua(to_pdf, language=None):
    # Create StructTreeRoot and descendants, allocate new xrefs as needed
    structtreeroot_xref = to_pdf.get_new_xref()
    parenttree_xref = to_pdf.get_new_xref()
    page_info_xrefs = []
    page_info_a_xrefs = []
    parenttree_kids_xrefs = []
    parenttree_kids_indirect_xrefs = []

    kids_cnt = ceil(to_pdf.page_count / 32)
    for _ in range(kids_cnt):
        kid_xref = to_pdf.get_new_xref()
        parenttree_kids_xrefs.append(kid_xref)

    # Parent tree contains a /Kids entry with a list of xrefs, that each contain
    # a list of xrefs (limited to 32 per), and each entry in that list of list
    # of xrefs contains a single reference that points to the page info xref.
    for idx, page in enumerate(to_pdf):
        page_info_xref = to_pdf.get_new_xref()
        page_info_xrefs.append(page_info_xref)

        page_info_a_xref = to_pdf.get_new_xref()
        page_info_a_xrefs.append(page_info_a_xref)

        parenttree_kids_indirect_xref = to_pdf.get_new_xref()
        parenttree_kids_indirect_xrefs.append(parenttree_kids_indirect_xref)


    for idx in range(kids_cnt):
        start = idx*32
        stop = (idx+1)*31
        if stop > to_pdf.page_count:
            stop = to_pdf.page_count- 1

        s = """<<
  /Limits [ %d %d ]
""" % (start, stop - 1)
        s += '  /Nums [ '

        for pidx in range(start, stop):
            s += '%d %d 0 R ' % (pidx, parenttree_kids_indirect_xrefs[pidx])

            if idx % 7 == 0:
                s = s[:-1] + '\n' + '      '

        s += ']\n>>'

        to_pdf.update_object(parenttree_kids_xrefs[idx], s)


    for idx, page in enumerate(to_pdf):
        intrect = tuple([int(x) for x in page.rect])

        s = """<<
  /BBox [ %d %d %d %d ]
  /InlineAlign /Center
  /O /Layout
  /Placement /Block
>>
""" % intrect
        to_pdf.update_object(page_info_a_xrefs[idx], s)

        s = """ <<
  /A %d 0 R
  /K 0
  /P %d 0 R
  /Pg %d 0 R
  /S /Figure
>>""" % (page_info_a_xrefs[idx], structtreeroot_xref, page.xref)

        to_pdf.update_object(page_info_xrefs[idx], s)


    for idx, page in enumerate(to_pdf):
        s = '[ %d 0 R ]' % page_info_a_xrefs[idx]
        to_pdf.update_object(parenttree_kids_indirect_xrefs[idx], s)


    K = '  /Kids [ '
    for idx in range(kids_cnt):
        K += '%d 0 R ' % parenttree_kids_xrefs[idx]

        if idx % 7 == 0:
            K = K[:-1] + '\n' + '      '

    K += ']'
    s = """<<
%s
>>
""" % K

    to_pdf.update_object(parenttree_xref, s)

    K = '  /K [ '
    for idx, xref in enumerate(page_info_xrefs):
        K += '%d 0 R ' % xref

        if idx % 7 == 0:
            K = K[:-1] + '\n' + '      '

    K += ']'

    to_pdf.update_object(structtreeroot_xref, """
<<
""" + K + """
  /Type /StructTreeRoot
  /ParentTree %d 0 R
>>
""" % parenttree_xref)

    #  TODO? /ClassMap 1006 0 R
    #  TODO? /ParentTreeNextKey 198


    # Update pages, add back xrefs
    for idx, page in enumerate(to_pdf):
        page_data = to_pdf.xref_object(page.xref)
        page_data = page_data[:-2]

        page_data += """
  /StructParents %d
""" % idx

        page_data += """
  /CropBox [ 0 0 %.1f %.1f ]
""" % (page.rect[2], page.rect[3])

        page_data += """
  /Rotate 0
"""
        page_data += """
  /Tabs /S
"""
        page_data += '>>'
        to_pdf.update_object(page.xref, page_data)

    catalogxref = to_pdf.pdf_catalog()
    s = to_pdf.xref_object(to_pdf.pdf_catalog())
    s = s[:-2]
    s += """
  /ViewerPreferences <<
    /FitWindow true
    /DisplayDocTitle true
  >>
"""
    if language:
        s += """
  /Lang (%s)
""" % language

    s += """
  /MarkInfo <<
    /Marked true
  >>
"""
    s += """
  /StructTreeRoot %d 0 R
""" % structtreeroot_xref

    s += '>>'
    to_pdf.update_object(catalogxref, s)


def write_metadata(from_pdf, to_pdf, extra_metadata):
    """
    Write document and XMP metadata.

    Args:

    * from_pdf (fitz.Document or None): metadata to copy from input PDF, can be omitted
    * to_pdf: (fitz.Document): PDF to write metadata to
    * extra_metadata (dict): dictionary with extra metadata values

    Allowed values for extra_metadata:

    * 'url'
    * 'title'
    * 'author'
    * 'creator'
    * 'subject'
    * 'creatortool'
    * 'language' (can be a list)
    """
    doc_md = from_pdf.metadata if from_pdf is not None else {}

    doc_md['producer'] = PRODUCER

    if 'url' in extra_metadata:
        doc_md['keywords'] = extra_metadata['url']
    if 'title' in extra_metadata:
        doc_md['title'] = extra_metadata['title']
    if 'author' in extra_metadata:
        doc_md['author'] = extra_metadata['author']
    if 'creator' in extra_metadata:
        doc_md['creator'] = extra_metadata['creator']
    if 'subject' in extra_metadata:
        doc_md['subject'] = extra_metadata['subject']

    current_time = 'D:' + datetime.utcnow().strftime('%Y%m%d%H%M%SZ')
    if from_pdf is not None:
        doc_md['creationDate'] = from_pdf.metadata['creationDate']
    else:
        doc_md['creationDate'] = current_time
    doc_md['modDate'] = current_time

    # Set PDF basic metadata
    to_pdf.set_metadata(doc_md)

    have_xmlmeta = (from_pdf is not None) and (from_pdf.xref_xml_metadata() > 0)
    if have_xmlmeta:
        xml_xref = from_pdf.xref_xml_metadata()

        # Just copy the existing XML, perform no validity checks
        xml_bytes = from_pdf.xref_stream(xml_xref)
        to_pdf.set_xml_metadata(xml_bytes.decode('utf-8'))
    else:
        current_time = datetime.utcnow().isoformat(timespec='seconds') + 'Z'

        stream='''<?xpacket begin="" id="W5M0MpCehiHzreSzNTczkc9d"?>
        <x:xmpmeta xmlns:x="adobe:ns:meta/">
          <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
            <rdf:Description rdf:about="" xmlns:xmp="http://ns.adobe.com/xap/1.0/">
              <xmp:CreateDate>{createdate}</xmp:CreateDate>
              <xmp:MetadataDate>{metadatadate}</xmp:MetadataDate>
              <xmp:ModifyDate>{modifydate}</xmp:ModifyDate>
              <xmp:CreatorTool>{creatortool}</xmp:CreatorTool>'''.format(creatortool=xmlescape(extra_metadata.get('creatortool', PRODUCER)),
           createdate=current_time, metadatadate=current_time,
           modifydate=current_time)

        stream += '''
            </rdf:Description>
            <rdf:Description rdf:about="" xmlns:pdf="http://ns.adobe.com/pdf/1.3/">'''

        if 'url' in extra_metadata:
            stream += '''
              <pdf:Keywords>{keywords}</pdf:Keywords>'''.format(keywords=xmlescape(extra_metadata['url']))

        stream += '''
              <pdf:Producer>{producer}</pdf:Producer>'''.format(producer=xmlescape(PRODUCER))

        stream += '''
            </rdf:Description>
            <rdf:Description rdf:about="" xmlns:dc="http://purl.org/dc/elements/1.1/">'''

        if extra_metadata.get('title'):
            stream += '''
              <dc:title>
                <rdf:Alt>
                  <rdf:li xml:lang="x-default">{title}</rdf:li>
                </rdf:Alt>
              </dc:title>'''.format(title=xmlescape(extra_metadata.get('title')))

        # "An entity responsible for making the resource."
        # https://www.dublincore.org/specifications/dublin-core/dcmi-terms/#http://purl.org/dc/terms/creator
        # So should be author...
        if extra_metadata.get('author'):
            stream += '''
              <dc:creator>
                <rdf:Seq>
                  <rdf:li>{author}</rdf:li>
                </rdf:Seq>
              </dc:creator>'''.format(author=xmlescape(extra_metadata.get('author')))

        # TODO: Support multiple languages here?

        if extra_metadata.get('language'):
        # Empty language field means unknown language
            stream += '''
              <dc:language>
                <rdf:Bag>'''

            for language in extra_metadata.get('language', []):
                stream += '''
                  <rdf:li>{language}</rdf:li>'''.format(language=xmlescape(language))

            stream += '''
                </rdf:Bag>
              </dc:language>'''

        stream += '''
            </rdf:Description>
            <rdf:Description rdf:about="" xmlns:pdfaid="http://www.aiim.org/pdfa/ns/id/">
              <pdfaid:part>3</pdfaid:part>
              <pdfaid:conformance>B</pdfaid:conformance>
            </rdf:Description>
          </rdf:RDF>
        </x:xmpmeta>
        <?xpacket end="r"?>'''

        to_pdf.set_xml_metadata(stream)



# archive-pdf-tools
# Copyright (C) 2020-2021, Internet Archive
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Merlijn Boris Wolf Wajer <merlijn@archive.org>

import sys
from os import close, remove

from glob import glob
from tempfile import mkstemp
import subprocess
from time import time

import warnings

from PIL import Image, ImageOps
from skimage.filters import threshold_local, threshold_otsu
from skimage.restoration import denoise_tv_bregman, estimate_sigma

from scipy import ndimage
import numpy as np

from optimiser import optimise_gray, optimise_rgb, optimise_gray2, optimise_rgb2, fast_mask_denoise
from sauvola import binarise_sauvola

import fitz

fitz.TOOLS.set_icc(True) # For good measure, not required

from internetarchivepdf.jpeg2000 import encode_jpeg2000
from internetarchivepdf.const import (RECODE_RUNTIME_WARNING_TOO_SMALL_TO_DOWNSAMPLE, COMPRESSOR_JPEG,
        COMPRESSOR_JPEG2000, DENOISE_NONE, DENOISE_FAST, DENOISE_BREGMAN)


"""
"""

# skimage throws useless UserWarnings in various functions
def mean_estimate_sigma(arr):
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        return np.mean(estimate_sigma(arr))


def threshold_image(img, dpi, k=0.34):
    """
    Perform Sauvola binarisation on the given image

    Args:

    * img (np.ndarray): input image array
    * dpi (int): dpi for Sauvola, used to calculate window size if not None
    * k (float): k parameter, defaults to 0.34

    Returns binarised numpy.ndarray
    """
    window_size = 51

    if dpi is not None:
        window_size = int(dpi / 4)
        if window_size % 2 == 0:
            window_size += 1

    h, w = img.shape
    out_img = np.ndarray(img.shape, dtype=np.bool)
    out_img = np.reshape(out_img, w*h)
    in_img = np.reshape(img, w*h)

    binarise_sauvola(in_img, out_img, w, h, window_size, window_size, k, 128)
    out_img = np.reshape(out_img, (h, w))
    # TODO: optimise this, we can do it in binarise_sauvola
    out_img = np.invert(out_img)

    return out_img


def denoise_bregman(binary_img):
    """
    Denoise a binary numpy array using Bregman total variation denoising

    Args:

    * binary_img (np.array): input array

    Returns the denoised array
    """
    thresf = np.array(binary_img, dtype=np.float32)
    #denoise = denoise_tv_bregman(thresf, weight=0.25)
    denoise = denoise_tv_bregman(thresf, weight=1.)

    #denoise = denoise > 0.6
    denoise = denoise > 0.4  # XXX: 0.4?
    denoise = np.array(denoise, dtype=np.bool)

    return denoise

# TODO: Rename, can be either foreground or background
def partial_blur(mask, img, sigma=5, mode=None):
    """
    Blur a part of the image 'img', where mask = 0.
    The actual values used by the blur are colours where mask = '1', effectively
    'erasing/blurring' parts of an image where mask = 0 with colours where mask = 1.

    At the end, restore all pixels from img where mask = 1.
    """
    maskf = np.array(mask, dtype=np.float32)

    if mode == 'RGB' or mode == 'RGBA':
        in_r = img[:, :, 0] * maskf
        in_g = img[:, :, 1] * maskf
        in_b = img[:, :, 2] * maskf
        filter_r = ndimage.filters.gaussian_filter(in_r, sigma = sigma)
        filter_g = ndimage.filters.gaussian_filter(in_g, sigma = sigma)
        filter_b = ndimage.filters.gaussian_filter(in_b, sigma = sigma)
    else:
        imgf = np.copy(img)
        imgf = np.array(imgf, dtype=np.float32)
        filter = ndimage.filters.gaussian_filter(imgf * maskf, sigma = sigma)

    weights = ndimage.filters.gaussian_filter(maskf, sigma = sigma)

    if mode == 'RGB' or mode == 'RGBA':
        filter_r /= weights + 0.00001
        filter_g /= weights + 0.00001
        filter_b /= weights + 0.00001

        newimg = np.copy(img)
        newimg[:, :, 0] = filter_r
        newimg[:, :, 1] = filter_g
        newimg[:, :, 2] = filter_b
    else:
        filter /= weights + 0.00001
        newimg = np.array(filter, dtype=np.uint8)

    newimg[mask] = img[mask]

    return newimg


def partial_boxblur(mask, fg, size=5, mode=None):
    maskf = np.array(mask, dtype=np.float32)

    if mode == 'RGB' or mode == 'RGBA':
        in_r = fg[:, :, 0] * maskf
        in_g = fg[:, :, 1] * maskf
        in_b = fg[:, :, 2] * maskf
        filter_r = ndimage.uniform_filter(in_r, size = size)
        filter_g = ndimage.uniform_filter(in_g, size = size)
        filter_b = ndimage.uniform_filter(in_b, size = size)
    else:
        fgf = np.copy(fg)
        fgf = np.array(fgf, dtype=np.float32)
        filter = ndimage.uniform_filter(fgf * maskf, size = size)

    weights = ndimage.uniform_filter(maskf, size = size)

    if mode == 'RGB' or mode == 'RGBA':
        filter_r /= weights + 0.00001
        filter_g /= weights + 0.00001
        filter_b /= weights + 0.00001

        newfg = np.copy(fg)
        newfg[:, :, 0] = filter_r
        newfg[:, :, 1] = filter_g
        newfg[:, :, 2] = filter_b
    else:
        filter /= weights + 0.00001
        newfg = np.array(filter, dtype=np.uint8)

    newfg[mask] = fg[mask]

    return newfg


def create_hocr_mask(img, mask_arr, hocr_word_data, downsample=None, dpi=None, timing_data=None):
    image_width, image_height = img.size
    np_img = np.array(img)

    t = time()

    for paragraph in hocr_word_data:
        for line in paragraph['lines']:
            coords = line['bbox']

            line_text = ' '.join([word['text'] for word in line['words']])
            line_confs = [word['confidence'] for word in line['words']]
            line_conf = sum(line_confs) / len(line_confs)

            if line_text.strip() == '' or line_conf < 20:
                continue

            if downsample is not None:
                coords = [int(x/downsample) for x in coords]
            else:
                coords = [int(x) for x in coords]

            left, top, right, bottom = coords
            # This can happen if we downsample and round to int
            if left == right or top == bottom:
                continue

            if (left >= right) or (top >= bottom):
                print('Invalid bounding box: (%d, %d, %d, %d)' % (left, top, right, bottom), file=sys.stderr)
                continue

            if (left < 0) or (right > image_width) or (top < 0) or (bottom > image_height):
                print('Invalid bounding box outside image: (%d, %d, %d, %d)' % (left, top, right, bottom), file=sys.stderr)
                continue

            np_lineimg = np_img[top:bottom,left:right]
            # Simple grayscale invert
            np_lineimg_invert = 255 - np.copy(np_lineimg)

            # XXX: If you tweak k, you must tweak the various ratio and sigma's
            # based on the test images
            k = 0.1
            thres = threshold_image(np_lineimg, dpi, k)
            ones = np.count_nonzero(thres)
            zero = np_lineimg.size - ones
            ratio = (ones/(zero+ones))

            thres_invert = threshold_image(np_lineimg_invert, dpi, k)
            ones_i = np.count_nonzero(thres_invert)
            zero_i = np_lineimg.size - ones_i
            inv_ratio = (ones_i/(zero_i+ones_i))

            if ratio < 0.3 or inv_ratio < 0.3:
                th = None

                perc_larger = 0.
                if inv_ratio != 0.0:
                    perc_larger = (ratio / inv_ratio) * 100

                if inv_ratio > 0.2 and ratio < 0.2:
                    th = thres
                else:
                    # mean_estimate_sigma is expensive, so let's only do it if
                    # we need to

                    ratio_sigma = mean_estimate_sigma(thres)
                    inv_ratio_sigma = mean_estimate_sigma(thres_invert)


                    # Prefer ratio over inv_ratio by a bit
                    if inv_ratio < 0.3 and inv_ratio < ratio and \
                    (inv_ratio_sigma < ratio_sigma or \
                    (ratio_sigma < 0.1 and inv_ratio_sigma < 0.1)):
                        th = thres_invert
                    elif ratio < 0.2:
                        th = thres

                if th is not None:
                    mask_arr[top:bottom, left:right] = th


    if timing_data is not None:
        timing_data.append(('hocr_mask_gen', time() - t))


def estimate_noise(imgf):
    #sigma_est = mean_estimate_sigma(imgf)
    #return sigma_est

    # We do this only on a part of the image, because it's accurate enough wrt
    # noise estimation (definitely for camera noise estimation since that's
    # everywhere in the image, and it's quite a bit faster this way).
    h, w = imgf.shape
    MUL = 4
    hs = int(h/2 - h/MUL)
    he = int(h/2 + h/MUL)
    ws = int(w/2 - w/MUL)
    we = int(w/2 + w/MUL)

    # Really small image?
    if he == 0 or we == 0:
        hs = 0
        he = h
        ws = 0
        we = w

    sigma_est = mean_estimate_sigma(imgf[hs:he, ws:we])

    return sigma_est



def create_threshold_mask(mask_arr, imgf, dpi=None, denoise_mask=None, timing_data=None):
    # We don't apply any of these blurs to the hOCR mask, we want that as
    # sharp as possible.

    t = time()
    sigma_est = estimate_noise(imgf)

    if timing_data is not None:
        timing_data.append(('est_1', time() - t))
    if sigma_est > 1.0:
        t = time()
        imgf = ndimage.filters.gaussian_filter(imgf, sigma=sigma_est*0.1)
        if timing_data is not None:
            timing_data.append(('blur_1', time() - t))

        #t = time()
        #n_sigma_est = mean_estimate_sigma(imgf)
        #time_data.append(('est_2', time() - t))
        #if sigma_est > 1.0 and n_sigma_est > 1.0:
            #    t = time()
        #    imgf = ndimage.filters.gaussian_filter(imgf, sigma=sigma_est*0.5)
        #    print('Going for second blur: n_sigma_est:',n_sigma_est)
        #    time_data.append(('blur_2', time() - t))

    t = time()
    thres_arr = threshold_image(imgf.astype(np.uint8), dpi)
    if timing_data is not None:
        timing_data.append(('threshold', time() - t))

    mask_arr |= thres_arr


# TODO: Reduce amount of memory active at one given point (keep less images in
# memory, write to disk sooner, etc), careful with numpy <-> PIL conversions
def create_mrc_hocr_components(image, hocr_word_data,
                               dpi=None,
                               downsample=None,
                               bg_downsample=None,
                               fg_downsample=None,
                               denoise_mask=None, timing_data=None,
                               errors=None):
    """
    Create the MRC components: mask, foreground and background

    Args:

    * image (PIL.Image): Image to be decomposed
    * hocr_word_data: OCR data about found text on the page
    * downsample (int): factor by which the OCR data is to be downsampled
    * bg_downsample (int): if the background image should be downscaled
    * denoise_mask (bool): Whether to denoise the image if it is deemed too
      noisy
    * timing_data: Optional timing data to log individual timing data to.
    * errors: Optional argument (of type set) with encountered runtime errors

    Returns a tuple of the components, as numpy arrays: (mask, foreground,
    background)
    """
    grayimg = image
    if image.mode != 'L':
        t = time()
        grayimg = image.convert('L')
        if timing_data is not None:
            timing_data.append(('grey_conversion', time() - t))

    width_, height_ = image.size

    mask_arr = np.array(Image.new('1', image.size))

    # Modifies mask_arr in place
    create_hocr_mask(grayimg, mask_arr, hocr_word_data, downsample=downsample,
                     dpi=dpi, timing_data=timing_data)
    grayimgf = np.array(grayimg, dtype=np.float32)

    MIX_THRESHOLD = True
    if MIX_THRESHOLD:
        # XXX: this nukes the hocr threshold, testing only
        # mask_arr = np.zeros(mask_arr.shape, dtype=np.bool)

        # Modifies mask_arr in place
        create_threshold_mask(mask_arr, grayimgf, dpi=dpi,
                              denoise_mask=denoise_mask,
                              timing_data=timing_data)

    if denoise_mask != DENOISE_NONE:
        t = time()
        if denoise_mask == DENOISE_FAST:
            # XXX: We could make the mincnt parameter take the dpi into account
            fast_mask_denoise(mask_arr, width_, height_, 4, 2)
            if timing_data is not None:
                timing_data.append(('fast_denoise', time() - t))
        elif denoise_mask == DENOISE_BREGMAN:
            mask_arr = denoise_bregman(mask_arr)
            if timing_data is not None:
                timing_data.append(('denoise', time() - t))
        else:
            raise ValueError('Invalid denoise option:', denoise_mask)


    yield mask_arr

    image_arr = np.array(image)

    t = time()
    # Take foreground pixels and optimise the image by making the surrounding
    # pixels like the foreground, allowing for more optimal compression (and
    # higher quality foreground pixels as a result)
    if image.mode == 'L':
        foreground_arr = optimise_gray2(mask_arr, image_arr, width_, height_, 3)
    else:
        foreground_arr = optimise_rgb2(mask_arr, image_arr, width_, height_, 3)
    if timing_data is not None:
        # The name fg_partial_blur is kept for backwards compatibility
        timing_data.append(('fg_partial_blur', time() - t))

    if fg_downsample is not None:
        t = time()
        image2 = Image.fromarray(foreground_arr)
        w, h = image2.size
        w_downsample = int(w / fg_downsample)
        h_downsample = int(h / fg_downsample)
        if w_downsample > 0 and h_downsample > 0:
            image2.thumbnail((w_downsample, h_downsample))
            foreground_arr = np.array(image2)
        else:
            if errors is not None:
                errors.add(RECODE_RUNTIME_WARNING_TOO_SMALL_TO_DOWNSAMPLE)

        if timing_data is not None:
            timing_data.append(('fg_downsample', time() - t))

    yield foreground_arr
    foreground_arr = None

    mask_inv = mask_arr ^ np.ones(mask_arr.shape, dtype=bool)

    t = time()
    # Take background pixels and optimise the image by placing them where the
    # foreground pixels are thought to be, this has the effect of reducing
    # compression artifacts (thus improving quality) and at the same time making
    # the image easier to compress (smaller file size)
    if image.mode == 'L':
        background_arr = optimise_gray2(mask_inv, image_arr, width_, height_, 10)
    else:
        background_arr = optimise_rgb2(mask_inv, image_arr, width_, height_, 10)
    if timing_data is not None:
        # The name bg_partial_blur is kept for backwards compatibility
        timing_data.append(('bg_partial_blur', time() - t))

    if bg_downsample is not None:
        t = time()
        image2 = Image.fromarray(background_arr)
        w, h = image2.size
        w_downsample = int(w / bg_downsample)
        h_downsample = int(h / bg_downsample)
        if w_downsample > 0 and h_downsample > 0:
            image2.thumbnail((w_downsample, h_downsample))
            background_arr = np.array(image2)
        else:
            if errors is not None:
                errors.add(RECODE_RUNTIME_WARNING_TOO_SMALL_TO_DOWNSAMPLE)

        if timing_data is not None:
            timing_data.append(('bg_downsample', time() - t))

    yield background_arr
    return


def encode_mrc_mask(np_mask, tmp_dir=None, jbig2=True, embedded_jbig2=False,
                    timing_data=None, debug=False):
    """
    Encode mask image either to JBIG2 or PNG.

    Args:

    * np_mask (numpy.array): Mask image array
    * tmp_dir (str): path the temporary directory to write images to
    * jbig2 (bool): Whether to encode to JBIG2 or PNG
    * embedded_jbig2 (bool): Whether to encode to JBIG2 with or without header
    * timing_data (optional): Add time information to timing_data structure

    Returns a tuple: (str, str) where the first entry is the jbig2
    path, if any, the second is the png path.
    """
    t = time()
    mask = Image.fromarray(np_mask)

    fd, mask_img_png = mkstemp(prefix='mask', suffix='.png', dir=tmp_dir)
    close(fd)
    if jbig2:
        fd, mask_img_jbig2 = mkstemp(prefix='mask', suffix='.jbig2', dir=tmp_dir)
        close(fd)

    mask.save(mask_img_png, compress_level=0)

    if jbig2:
        args = ['jbig2', mask_img_png]
        if embedded_jbig2:
            args = ['jbig2', '-p', mask_img_png]

        if debug:
            print('check_output: %s' % args, file=sys.stderr)

        out = subprocess.check_output(args)
        fp= open(mask_img_jbig2, 'wb+')
        fp.write(out)
        fp.close()

    if timing_data is not None:
        timing_data.append(('mask_jbig2', time()-t))

    if jbig2:
        return mask_img_jbig2, mask_img_png
    else:
        return None, mask_img_png


def encode_mrc_img(np_img, img_compression_flags, imgtype=None, tmp_dir=None,
        jpeg2000_implementation=None, mrc_image_format=None, timing_data=None,
        threads=False,
        debug=False):
    """
    Encode image as JPEG2000 or JPEG, with the provided compression settings
    and JPEG2000/JPEG encoder.

    Args:

    * np_img (numpy.array): Image array
    * img_compression_flags (str): Compression flags
    * imgtype (str: 'bg' or 'fg'
    * tmp_dir (str): path the temporary directory to write images to
    * jpeg2000_implementation (str): What JPEG2000 implementation to use
    * mrc_image_format (str): What image format to produce
    * timing_data (optional): Add time information to timing_data structure
    * debug (bool, optional): Write debug info to stderr

    Returns the filepath to the JPEG2000 image
    """
    t = time()
    if imgtype not in ('bg', 'fg'):
        raise ValueError('imgtype should be \'bg\' or \'fg\'')

    # Create background
    if mrc_image_format == COMPRESSOR_JPEG:
        fd, img_tiff = mkstemp(prefix=imgtype, suffix='.jpg', dir=tmp_dir)
        close(fd)

    fd, img_jp2 = mkstemp(prefix=imgtype, suffix='.jp2', dir=tmp_dir)
    close(fd)
    remove(img_jp2) # XXX: Kakadu doesn't want the file to exist, so what are
                       # we even doing

    img = Image.fromarray(np_img)

    if mrc_image_format == COMPRESSOR_JPEG:
        img.save(img_tiff, quality=100)


        args = ['jpegoptim'] + img_compression_flags + [img_tiff, '--stdout']
        if debug:
            print('check_output: %s' % args, file=sys.stderr)
        output = subprocess.check_output(args)
        tmpfd=open(img_jp2, 'bw+') # XXX: FIXME: this defeats the point of a tmpfile
        tmpfd.write(output)
        tmpfd.close()
    else:
        encode_jpeg2000(img, img_jp2, jpeg2000_implementation,
                        img_compression_flags, imgtype=imgtype,
                        threads=threads, debug=debug)


    if timing_data is not None:
        timing_data.append(('%s_jp2' % imgtype, time()-t))

    return img_jp2


def encode_mrc_background(np_bg, bg_compression_flags, tmp_dir=None,
        jpeg2000_implementation=None, mrc_image_format=None, timing_data=None,
        threads=None, debug=False):
    """
    Encode background image as JPEG2000, with the provided compression settings
    and JPEG2000 encoder.

    Args:

    * np_bg (numpy.array): Background image array
    * bg_compression_flags (str): Compression flags
    * tmp_dir (str): path the temporary directory to write images to
    * jpeg2000_implementation (str): What JPEG2000 implementation to use
    * mrc_image_format (str): What image format to produce
    * timing_data (optional): Add time information to timing_data structure

    Returns the filepath to the JPEG2000 background image
    """
    return encode_mrc_img(np_bg, bg_compression_flags, 'bg', tmp_dir=tmp_dir,
            jpeg2000_implementation=jpeg2000_implementation,
            mrc_image_format=mrc_image_format, timing_data=timing_data,
            threads=threads,
            debug=debug)


def encode_mrc_foreground(np_fg, fg_compression_flags, tmp_dir=None,
        jpeg2000_implementation=None, mrc_image_format=None, timing_data=None,
        threads=False, debug=False):
    """
    Encode foreground image as JPEG2000, with the provided compression settings
    and JPEG2000 encoder.

    Args:

    * np_bg (numpy.array): Foreground image array
    * fg_compression_flags (str): Compression flags
    * tmp_dir (str): path the temporary directory to write images to
    * jpeg2000_implementation (str): What JPEG2000 implementation to use
    * mrc_image_format (str): What image format to produce
    * timing_data (optional): Add time information to timing_data structure

    Returns the filepath to the JPEG2000 foreground image
    """
    return encode_mrc_img(np_fg, fg_compression_flags, 'fg', tmp_dir=tmp_dir,
            jpeg2000_implementation=jpeg2000_implementation,
            mrc_image_format=mrc_image_format, timing_data=timing_data,
            threads=threads,
            debug=debug)


def encode_mrc_images(mrc_gen, bg_compression_flags=None, fg_compression_flags=None,
                      tmp_dir=None, jbig2=True, timing_data=None,
                      jpeg2000_implementation=None, mrc_image_format=None,
                      embedded_jbig2=False, threads=None, debug=False):
    mask_img_jbig2, mask_img_png = encode_mrc_mask(next(mrc_gen),
            tmp_dir=tmp_dir, jbig2=jbig2, embedded_jbig2=embedded_jbig2,
            timing_data=timing_data)

    np_fg = next(mrc_gen)
    fg_img_jp2 = encode_mrc_foreground(np_fg, fg_compression_flags, tmp_dir=tmp_dir,
                                       jpeg2000_implementation=jpeg2000_implementation,
                                       mrc_image_format=mrc_image_format,
                                       timing_data=timing_data, threads=threads, debug=debug)
    fg_h, fg_w = np_fg.shape[0:2]
    np_fg = None

    np_bg = next(mrc_gen)
    bg_img_jp2 = encode_mrc_background(np_bg, bg_compression_flags, tmp_dir=tmp_dir,
                                       jpeg2000_implementation=jpeg2000_implementation,
                                       mrc_image_format=mrc_image_format,
                                       timing_data=timing_data, threads=threads, debug=debug)
    bg_h, bg_w = np_bg.shape[0:2]
    np_bg = None

    # XXX: probably don't need this
    try:
        _ = next(mrc_gen)
    except StopIteration:
        pass

    if jbig2:
        remove(mask_img_png)

    if jbig2:
        return mask_img_jbig2, bg_img_jp2, (bg_w, bg_h), fg_img_jp2, (fg_w, fg_h)
    else:
        # Return PNG which mupdf will turn into ccitt with
        # save(..., deflate=True) until mupdf fixes their JBIG2 support
        #return mask_img_png, bg_img_jp2, fg_img_jp2
        return mask_img_png, bg_img_jp2, (bg_w, bg_h), fg_img_jp2, (fg_w, fg_h)
